/* Internet HTTP client
   Copyright (c) John L Washington, 1996-7, ka9qwww@wash.demon.co.uk.
   Some parts derived from telnet.c, which is Copyright 1991 Phil Karn, KA9Q.

   John Washington's moral rights under the Copyrights, Designs 
   and Patents Act are asserted.															
 */

/*
 *	$Id: httpcli.c 1.1 1997/10/04 20:52:22 Graham Exp $
 *	Mar 97	1.0	JLW	Original version
 *	Oct 97	1.1	GeK	Added Host: and User-Agent: headers
 */

/*
http url [options]

where options are one or more of:

    file=
        if no filename follows the '=', the file isn't stored
        and view is implied.

    force
        Even the header is stored.

    view
        The file content is displayed.  No pause after n lines,
        might add that later?
        The file is still stored, unless 'file=' is used.

    link
        View the links.  Turn off the 'view' option.  So,
        if you use it with 'file=', put it after that, not
        before, otherwise you'll view the whole file.
        This option is experimental and might be removed
        eventually.

    label
        Add a line at the start of the file, to show where this
        file was fetched from.

    head
        Use the HEAD form of command instead of GET.  This fetches
        only the header of the file.

Script file support

    http @myscript

This variant reads from the file 'myscript', so that a list of
URLs can be prepared offline.  Options can be set within the
script file, and/or on the line invoking the script (in the
latter case, the options are then applied to all lines within
the script).  Blank lines in the script are ignored.

*/

/*
TO DO:
(from Walter)
wb> I can't see why the session closes without a keystroke for a single file
wb> but it seems to do so.  Is it possible to build with debug on the module
wb> in question and run against the local site as both host and client.
wb> In the meantime, I would be inclined to drop the condition governing the
wb> keywait in dohttpcli.
*/

#include <ctype.h>
#include <stdio.h>
#ifdef  __TURBOC__
#include <io.h>
#include <fcntl.h>
#endif
#include "config.h"
#include "global.h"
#include "mbuf.h"
#include "socket.h"
#include "telnet.h"
#include "session.h"
#include "proc.h"
#include "tty.h"
#include "commands.h"
#include "netuser.h"

#ifdef HTTP

/*  GeK - what's sent in User-Agent header - see RFC2068 for acceptable contents  */
#define	USER_AGENT	"GremNOS-JW-httpcli/1.1"

static char dummyhost[] = "nohost.nowhere.nocom";


static int
dohttp1 (int argc, char* argv[],
    struct session* sp
)
{
    struct sockaddr_in fsocket;
    const char* url;
    unsigned int index;
    struct telnet tn;
    int ch;
    int s;
    char *cp;
    char* host;
    int got_host = 0;
    char* final_slash = NULLCHAR;
    char* filename = "index.htm";
    int x;
    FILE* fp = NULL;
    char buf [50];  /* For looking at lines in header, doesn't need to be long */
    size_t bufx;    /* Next free char in 'buf' */
        /* 'buf' and 'bufx' are also going to be used for temp storing links,
            in which case they may need to be longer */
    int capture = 0;    /* Set when capturing to file */
    int label = 0;      /* Set when adding url to front of file */
    int view = 0;
    int view_links = 0;
    int store = 1;
    int head = 0;
    int nl = 0;
    int addcr = 0;
    int content_html = 0;   /* Used when looking for links */
    int link = 0;
        /* Normally 'link' is 0.  While looking for '<a href="', 'link'
            steps thru the candidate string.  If it matches, 'link' is
            then set -ve while the link is stored in 'buf'
            */
    long bytecount = 0;

    /*
    log (sp->s, "HTTP: %s", argv[1]);   
    */
    
    cp = argv[1];
    if (*cp == '<')
    {               /* Strip off surrounding angle brackets, if any */
        size_t len = strlen (cp);
        if (cp [len - 1] == '>')
            cp [len - 1] = '\0';
        ++cp;
    }
    if (strnicmp (cp, "http:", 5) == 0)
        cp += 5;    /* Strip off 'http:', if any */
    if (strncmp (cp, "//", 2) == 0)
        cp += 2;    /* Strip off '//' if any */
    host = cp;      /* Start of host name found */
    fsocket.sin_port = 80;  /* Default port is 80 */
    url = "";
    for (; *cp != '\0'; ++cp)
    {
        ch = *cp;
        if (ch == '/')
        {
            if (!got_host)
            {
                *cp = '\0';     /* End of host name */
                url = cp + 1;
                got_host = 1;
            }
            final_slash = cp;
        }
        else if (ch == ':')
        {
            /*> Haven't bothered with check for multiple ':' */
            fsocket.sin_port = atoi (cp+1);
            *cp = '\0';     /* End of host name */
        }
    }
    if (final_slash != NULLCHAR && final_slash[1] != '\0')
    {
        filename = final_slash + 1;
            /* Unless there's a 'file=' param supplied, use the
                last part of the URL as the destination filename */
    }

                /* Process optional extra command args */
    for (x = 2; x < argc; ++x)
    {
        cp = argv [x];
        if (strnicmp (cp, "file=", 5) == 0)
        {
            filename = cp + 5;
            if (filename [0] == '\0')
            {
                store = 0;      /* 'file=' with no name views only */
                view = 1;
                view_links = 0;
            }
        }
        else if (strnicmp (cp, "force", 5) == 0)
            capture = 1;    /* Force capture immediately */
        else if (strnicmp (cp, "view", 4) == 0)
            view = 1;    
        else if (strnicmp (cp, "link", 4) == 0)
        {
            view = 0;    
            view_links = 1;    
        }
        else if (strnicmp (cp, "label", 5) == 0)
        {
            label = 1;    
            if (cp [5] == '=')
            {
                /*> Compare cp+6 with url */
            }
        }
        else if (strnicmp (cp, "head", 4) == 0)
        {
            head = 1;
            capture = 1;    /* Force capture immediately */
        }
    }

                /* Scan the filename for Unix '/' and change any to msdos '\' */
    for (cp = filename; ; ++cp)
    {
        ch = *cp;
        if (ch == '\0')
            break;
        if (ch == '/')
            ch = '\\';  /* Change to dos convention */
    }

    if (strcmp (sp->name, dummyhost) != 0)
    {
        tprintf ("Internal error");
        return 1;
    }

    #ifdef xxx
    {
    void debug (const char* fstr, ...);
    time_t nowTime;
    struct tm * tblock;
    char str[40];
    time_t timeDiff;
    static int diagctr;
      nowTime = time(NULL);
      tblock = localtime(&nowTime);
      strncpy(str, asctime(tblock), sizeof(str)-1);
      str[sizeof(str)-1] = '\0';
      if (str[strlen(str)-1] == '\n')
         str[strlen(str)-1] = '\0';
    tprintf ((diagctr % 2) ? "\n" : "\r");
    tprintf ("%d %s ", diagctr++, str);
                    usflush (Curproc->output);
                    alert (sp->Display, 1);
                    psignal (sp->Display, 0);
                    pwait (NULL);
      alarm(1000);
      pwait(Curproc);
      return 0;
    }
    #endif

    free (sp->name);
    sp->name = strdup (host);

    fsocket.sin_family = AF_INET;

    tprintf ("%s:%d /", host, fsocket.sin_port);
    tprintf ("%s\n", url);
    if (filename [0] != '\0')
        tprintf ("%s\n", filename);
    tprintf ("Resolving %s... ", sp->name);
    if ((fsocket.sin_addr.s_addr = resolve (sp->name)) == 0)
    {
        tprintf (Badhost, sp->name);
        log (sp->s, Badhost, sp->name);
        free (sp->name);
        sp->name = strdup (dummyhost);
        return 1;
    }
    if ((sp->s = socket (AF_INET, SOCK_STREAM, 0)) == -1)
    {
        tprintf ("Can't create socket\n");
        free (sp->name);
        sp->name = strdup (dummyhost);
        return 1;
    }

                /* Try to create the file */
    if (store)
    {
        fp = fopen (filename, "rb");
        if (fp != NULL)
        {
            fclose (fp);
            tprintf ("File %s exists", filename);
            ch = keywait ("Overwrite? (y/n)", 1);
                    /* 1 means flush input, might change this to 0? */
            if (tolower (ch) != 'y')
            {
                return 2;   /* Return code 1, but don't wait for keypress */
                            /*//> What do these return codes mean anyway?
                                In dohttp() I'm just copying the convention 
                                that seems to exist, 0 is good, 1 is bad, 
                                but I don't know where (or if) they're used. 
                                */
            }
        }
        fp = fopen (filename, "wb");
        /*//> Ought to check 'fp' here, trying to save space */
        if (label)
        {
            int port = fsocket.sin_port;
            char portstr [20] = "";
            if (port != 80)
                sprintf (portstr, ":%d", port);
            fprintf (fp, "<!-- http://%s%s/%s -->\r\n", host, portstr, url);
        }
    }

    /*//> Some of this stuff is copied from telnet.c and perhaps isn't needed? */
    index = (unsigned int) (sp - Sessions);
    memset ((char *)&tn, 0, sizeof tn);
    tn.eolmode = 0;
    tn.session = sp;    /* Upward pointer */
    sp->cb.telnet = &tn;    /* Downward pointer */
    sockmode (sp->s, SOCK_ASCII); /* Default to ascii mode */

    tprintf ("Trying %s...\n", psocket ((struct sockaddr *)&fsocket));
    if (connect (sp->s, (char*)&fsocket, SOCKSIZE) == -1)
    {
        if (fp != NULL)
            fclose (fp);
        tprintf ("HTTP session %u failed: %s errno %d\n",
            index, sockerr (sp->s), errno);

        return 1;
    }
    s = sp->s;
    tprintf ("HTTP session %u connected to %s\n", index, sp->name);
    log (s, "HTTP session %u connected to %s (%s)", index, sp->name, url);
    if (store && fp != NULL)
        log (s, "HTTP: %s", filename);

    sp->proc1 = NULLPROC;
        /*//> Not sure if this is needed */

    usprintf (s, "%s /%s HTTP/1.0\n", head ? "HEAD" : "GET", url);
    /*  GeK - send HTTP/1.1 Host: header for virtual servers sharing 1 IP addr  */
    usprintf (s, "Host: %s", host);
    if (fsocket.sin_port != 80)
        usprintf (s, ":%u", fsocket.sin_port);
    usputs (s, "\nUser-Agent: " USER_AGENT "\n\n");

                /* Process input from the connection */
    buf [0] = '\0';
    bufx = 0;
    while ((ch = recvchar (s)) != -1)
    {
        /*//> Is there a need to watch for and strip out \r?
          //> I've not seen any, but perhaps that's because
          //> so far I've only connected to Unix servers.
          //> Do any others put in \r\n, either in the header
          //> or in text files?
        */
        if (capture)
        {
            if (view)
                tputc (ch);
            else if (content_html)
            {
                if (view_links)
                {
                    if (link == 0)
                    {
                        if (ch == '<')
                            link = 1;
                    }
                    else if (link == -1)
                    {
                        /* Storing link */
                        /*//> for testing, it's not yet stored, just displayed */
                        if (bufx == 0)
                            tputc ('\n');
                        bufx++;
                        if (ch == '"')
                            link = 0;
                        else
                            tputc (ch);
                    }
                    else
                    {
                        static const char* const ahref = "<a href=\"";
                        if (ch == '\0')
                        {
                            /* Don't think this can happen, but just to be safe */
                            link = 0;
                        }
                        else if (tolower(ch) == tolower(ahref [link]))
                        {
                            link++;
                            if (ahref [link] == '\0')
                            {
                                /* Found link */
                                link = -1;
                                bufx = 0;
                            }
                        }
                        else if (isspace (ch) && isspace (ahref [link-1]))
                        {
                            /* Skipping whitespace */
                        }
                        else
                            link = 0;
                    }
                }
            }
            if (fp != NULL)
            {
                if (addcr && ch == '\n')
                    fputc ('\r', fp);
                fputc ((char)ch, fp);
                bytecount++;
                if ((bytecount % 1000) == 0)
                {
                    tprintf ("\r%ld ", bytecount);
                    usflush (Curproc->output);
                    #ifdef TEXTWIN
                    alert (sp->Display, 1);
                    psignal (sp->Display, 0);
                    pwait (NULL);
                    #endif
                }
            }
        }
        else /* not capture */
        {
            tputc (ch);
            if (ch == '\n')
            {
                if (nl && store)
                {   /* double newline, so start of actual file */
                    capture = 1;
                    if (!addcr)
                    {
                        sockmode (s, SOCK_BINARY);
                        /*
                        log (s, "HTTP binary");
                        */
                    }
                }
                else if (bufx != 0)
                {
                    buf [bufx] = '\0';
                    log (s, "HTTP: %s", buf);
                    if (strnicmp (buf, "Content-type: ", 14) == 0)
                    {
                        if (strnicmp (buf + 14, "text/", 5) == 0)
                        {
                            if (strnicmp (buf + 19, "plain", 5) == 0)
                                addcr = 1;
                            else if (strnicmp (buf + 19, "html", 4) == 0)
                            {
                                addcr = 1;
                                content_html = 1;
                            }
                        }
                    }
                }
                nl = 1;
                bufx = 0;
                buf [0] = '\0';
            }
            else /* ch != '\n' */
            {
                nl = 0;
                if (bufx < sizeof buf - 1)
                {
                    buf[bufx++] = ch;
                }
            }
        } /* else not capture */
    }
    /* A close was received from the remote host.
     * Notify the user and wait for a response
     * from the user before freeing the session.
     */

    cp = sockerr (s);
    if (fp != NULL)
    {
        fclose (fp);
        tprintf ("\r%ld ", bytecount);
    }
    tprintf ("\nHTTP session %u", index);
    tprintf (" closed: %s\n", cp != NULLCHAR ? cp : "EOF");
    close_s (sp->s);
    sp->s = -1;
    free (sp->name);
    sp->name = strdup (dummyhost);
    return 0;
} /* dohttp1 */

#define INBUFSIZ 150
    /* Max length of script file line, arbitrary choice */

/* Execute user HTTP command */
int
dohttpcli (int argc, char *argv[], void *p)
{
    struct session* sp;
    int rc;
    if (argc < 2)
        return 1;   /* Neither URL nor @file */

                /* Allocate a session descriptor */
    if ((sp = newsession (dummyhost, TELNET)) == NULLSESSION)
    {
        tprintf ("Too many sessions\n");
        return 1;
    }

    log (sp->s, "HTTP %s", argv[1]);    /*//> DELETE SOON? */
    if (argv[1][0] != '@')
    {
        rc = dohttp1 (argc, argv, sp);
    }
    else
    {
        /* Read script from file */
        FILE* fp;
        rc = 1;
        fp = fopen (argv[1] + 1, "r");  /*//> used to be "rb" */
        if (fp == NULL)
        {
            tprintf ("No script file '%s'\n", argv[1] + 1);
        }
        else
        {
            char* inbuf = mallocw (INBUFSIZ);
            while (fgets (inbuf + 1, INBUFSIZ - 1, fp) != NULLCHAR)
            {
                /* Parse each line of the script file, setting up argc/argv
                    style arguments and hand them to the main http code.
                    */
                /* The following code is slightly tricky.  The input buffer
                    is prefaced with a space, so that the code for stripping
                    leading spaces and setting up several arguments is
                    simplified.
                    */
                size_t len;
                char* cp;
                char* largv [10];
                    /* 10 should be plenty, in real life */
                int largc;
                inbuf [0] = ' ';
                len = strlen (inbuf + 1);
                tprintf ("%s", inbuf + 1);
                if (inbuf [len] == '\n')
                    inbuf [len] = '\0';
                cp = inbuf;
                log (sp->s, "HTTP%s", cp);    /*//>> DELETE SOON? */

                /* Break up the line into an array of arguments */
                largc = 1;
                for (; ; cp++)
                {
                    int ch = *cp;
                    if (ch == '\0')
                        break;
                    if (isspace (ch))
                    {
                        *cp = '\0';
                        cp++;
                        while (isspace (*cp))
                            cp++;   /* Multiple whitespace between args is ok */
                        if (*cp == '\0')
                            break;  /* Ignore trailing whitespace */
                        largv [largc] = cp;
                        largc++;
                        if (largc >= (sizeof largv) / (sizeof largv[0]))
                            break;  /*//> Ought to report over-complex line */
                    }
                }
                if (largc > 1)      /* Ignore blank lines */
                {
                    int x;
                    for (x = 2; x < argc; x++)
                    {
                        /* Append extra args */
                        largv [largc] = argv [x];
                        largc++;
                        if (largc >= (sizeof largv) / (sizeof largv[0]))
                            break;  /*//> Ought to report over-complex line */
                    }
                    rc = dohttp1 (largc, largv, sp);
                }
            }
            free (inbuf);
            fclose (fp);
        }
    }
    if (rc == 2)
        rc = 1;
    else
        keywait (NULLCHAR, 1);
    freesession (sp);
    return rc;
}

#endif /* ifdef HTTP */

