/*
 *	pop3grab.c - Grab messages via POP3 and place in local mail queue
 *
 *	This version uses POP3 and Demon's proprietory *env command
 *
 *	GeK		27/10/97
 *
 *	$Id: $
 */

#include <stdio.h>
#include <string.h>
#include "global.h"
#include "mbuf.h"
#include "socket.h"
#include "usock.h"
#include "session.h"
#include "proc.h"
#include "netuser.h"
#include "commands.h"
#include "tty.h"
#include "main.h"
#include "cmdparse.h"
#include "smtp.h"
#include "pc.h"			/*  GeK - for smartdrive_commit_cache()  */

#ifndef	IPPORT_POP3
#  define	IPPORT_POP3	110
#endif	/* ifndef IPPORT_POP3 */

#define	POP3_CONNECT_TIMEOUT	connect_wait_val	/*  connect timeout  */
#define	POP3_RESPONSE_TIMEOUT	120000UL			/*  2 minute response timeout  */
#undef	RUN_IN_SESSION

#define LINE_BUF_LEN	512		/*  RFC says this ...  */

/*
 *	Gets POP3 response into buffer.
 *
 *	Returns -1 on recv failure
 *	         0 on a + response
 *	         1 on a - response
 *	         2 otherwise
 */
static int
get_pop3_response (int s, char *buffer, int buf_len)
{
	char	*cp;

	alarm (POP3_RESPONSE_TIMEOUT);
	if (recvline (s, buffer, buf_len) > 0)
	{
		alarm (0L);			
		rip (buffer);

		/*  This should be handled elsewhere  */
		if (buffer[0] == '-')
			tprintf ("Unexpected failure: '%s'\n", buffer);

		return (buffer[0] == '+' ? 0 :
		        buffer[0] == '-' ? 1 :
		                           2);
	}

	cp = !run_timer (&Curproc->alarm) ? "timeout" : sockerr (s);
	alarm (0L);
	tprintf ("Receive failed: %s\n", cp != NULL ? cp : "");

	return (-1);
}

/*
 *	Sends POP3 command
 */
static int
send_pop3_command (int s, char *buffer)
{
	return (usputs (s, buffer));
} 

/*
 *	Reads response to Demon's proprietory *env command
 *
 *	Returns -1 on recv failure
 *	         0 otherwise.
 */
static int
process_env_response (int s, char *buffer, int buf_len, char **to, char **from)
{	
	int	line_count = 0;

	while (recvline (s, buffer, buf_len) > 0)
	{
		line_count++;
		
		/*  Dot transparency  */
		if (buffer[0] == '.')
		{
			char	*src, *dst;

			if (buffer[1] == '\n')
				return (0);
				
			src = dst = buffer;
			while ((*(src++) = *(++dst)) != '\0')
				;
		}

		rip (buffer);

		switch (line_count)
		{
			case 1:		/*  RD or OK  */
			case 2:		/*  line 2 is undefined (nodename)  */
				break;
			case 3:
				/*  From envelope  */
				*from = strdup (buffer);
				break;
			case 4:
				/*  To envelope  */
				*to = strdup (buffer);
				break;
			default:
				break;
		}
	}
	
	return (-1);
}

static int
retrieve_mail_to_file (int s, char *buffer, int buf_len, FILE *mail_file)
{
	char	*cp;
	int		last_had_eol = TRUE;
	int		byte_count;

	alarm (POP3_RESPONSE_TIMEOUT);
	while ((byte_count = recvline (s, buffer, buf_len)) > 0)
	{
		alarm (0L);			
		cp = buffer;
		if (last_had_eol && buffer[0] == '.')
		{
			if (buffer[1] == '\n')
				return (0);
			else
				cp++;	/*  Dot transparency  */
		}
		fputs (cp, mail_file);
		last_had_eol = (buffer[byte_count - 1] == '\n');
	}

	cp = !run_timer (&Curproc->alarm) ? "timeout" : sockerr (s);
	alarm (0L);
	tprintf ("Receive failed: %s\n", cp != NULL ? cp : "");

	return (-1);
}

/*  argv[1] = server name  */
/*  argv[2] = mailbox  */
/*  argv[3] = password  */
int
dopop3grab (int argc, char *argv[], void *p)
{
	struct sockaddr_in sock;
	char           *cp;
	char           *buf;
	int             s = -1;
#ifdef	RUN_IN_SESSION
	struct session *sp;
#endif
	int             result;
	int             success = FALSE;
	unsigned long   num_messages;
	unsigned long   total_bytes;
	unsigned long   message_idx;
	unsigned long   size;
	char           *to = NULL,
	               *from = NULL,
	               *uidl = NULL;

#ifdef	RUN_IN_SESSION
	/* Allocate a session descriptor */
	if ((sp = newsession (argv[0], TELNET)) == NULL)
	{
		tputs ("Too many sessions\n");
		keywait (NULL, 1);
		return (1);
	}
#endif

	buf = mallocw (LINE_BUF_LEN);

#ifdef	RUN_IN_SESSION
	sp->ttystate.echo = sp->ttystate.edit = 0;
	sp->flowmode = 1;
#endif
	sock.sin_family = AF_INET;
	sock.sin_port = IPPORT_POP3;

	tprintf ("Resolving %s ... ", argv[1]);

	if ((sock.sin_addr.s_addr = resolve ((char *) argv[1])) == 0)
	{
		tprintf ("Host %s unknown\n", argv[1]);
		goto abort_connection;
	}
	tprintf ("trying %s", psocket ((struct sockaddr *) &sock));
#ifdef	RUN_IN_SESSION
	if ((sp->s = s = socket (AF_INET, SOCK_STREAM, 0)) < 0)
#else
	if ((s = socket (AF_INET, SOCK_STREAM, 0)) < 0)
#endif
	{
		tputs("\nCan't create socket\n");
		goto abort_connection;
	}
	sockmode (s, SOCK_ASCII);

	alarm (POP3_CONNECT_TIMEOUT);
	if (connect (s, (char *)&sock, sizeof (sock)) < 0)
	{
		cp = !run_timer (&Curproc->alarm) ? "timeout" : sockerr (s);
		alarm (0L);
		tprintf ("\nConnect failed: %s\n", cp != NULL ? cp : "");
		goto abort_connection;
	}
	alarm (0L);

	tputc ('\n');

	/*  Eat opening banner  */
	result = get_pop3_response (s, buf, LINE_BUF_LEN);
	if (result != 0)
		goto send_quit;

	/*  Username  */
	sprintf (buf, "USER %s\n", argv[2]);
	send_pop3_command (s, buf);
	result = get_pop3_response (s, buf, LINE_BUF_LEN);
	if (result != 0)
		goto send_quit;

	/*  Password  */
	sprintf (buf, "PASS %s\n", argv[3]);
	send_pop3_command (s, buf);
	result = get_pop3_response (s, buf, LINE_BUF_LEN);
	if (result != 0)
		goto send_quit;

	tprintf ("Logged in; retrieving mail for %s ...\n", argv[2]);

	/*  Stats  */
	send_pop3_command (s, "STAT\n");
	result = get_pop3_response (s, buf, LINE_BUF_LEN);
	if (result != 0)
		goto send_quit;
	result = sscanf (buf, "+OK %lu %lu", &num_messages, &total_bytes);
	if (result != 2)
		goto send_quit;

	tprintf ("%s: %lu messages (%lu bytes)\n\n", argv[2], num_messages, total_bytes);

	for (message_idx = 1; message_idx <= num_messages; message_idx++)
	{
		long	mqueue_id;
		FILE	*mail_file;
		char	prefix[9];
		char	mail_filename[80];
		time_t	time_now;
		int		disk_error = FALSE;

		/*  Attempt *ENV first as it's the most volatile  */

		/*  *env  */
		sprintf (buf, "*ENV %lu\n", message_idx);
		send_pop3_command (s, buf);
		result = get_pop3_response (s, buf, LINE_BUF_LEN);
		if (result != 0)
			goto send_quit;
		result = process_env_response (s, buf, LINE_BUF_LEN, &to, &from);
		if (result != 0)
			goto send_quit;

		/*  Size  */
		sprintf (buf, "LIST %lu\n", message_idx);
		send_pop3_command (s, buf);
		result = get_pop3_response (s, buf, LINE_BUF_LEN);
		if (result != 0)
			goto send_quit;
		result = sscanf (buf, "+OK %lu %lu", &message_idx, &size);
		if (result != 2)
			goto send_quit;

		/*  UIDL  */
		sprintf (buf, "UIDL %lu\n", message_idx);
		send_pop3_command (s, buf);
		result = get_pop3_response (s, buf, LINE_BUF_LEN);
		if (result != 0)
			goto send_quit;
		cp = strchr (buf, ' ');
		if (cp == NULL)
			goto send_quit;
		cp = strchr (cp+1, ' ');
		if (cp == NULL)
			goto send_quit;
		uidl = strdup (cp+1);

		sprintf (buf, "RETR %lu\n", message_idx);
		send_pop3_command (s, buf);
		result = get_pop3_response (s, buf, LINE_BUF_LEN);
		if (result != 0)
			goto send_quit;

		/*  Get an ID for this message  */
		mqueue_id = get_msgid();

		sprintf (prefix, "%ld", mqueue_id);
		mlock (Mailqdir, prefix);

		/*  Wrk file  */
		sprintf (mail_filename, "%s/%s.wrk", Mailqdir, prefix);
		mail_file = fopen (mail_filename, WRITE_TEXT);
		if (mail_file == NULL)
			goto unlink_wrk;

		/*  Queue host  */
		fprintf (mail_file, "%s\n", Hostname);
		/*  From envelope  */
		fprintf (mail_file, "%s\n", from);
		/*  To envelope  */
		fprintf (mail_file, "%s\n", to);

		if (ferror (mail_file))
			disk_error = TRUE;

		if (fclose (mail_file) == EOF)
			disk_error = TRUE;
			
		if (disk_error)
			goto unlink_wrk;

		/*  Text file  */
		sprintf (mail_filename, "%s/%s.txt", Mailqdir, prefix);

		mail_file = fopen (mail_filename, WRITE_TEXT);
		if (mail_file == NULL)
			goto unlink_txt;

		time_now = time (NULL);
		fprintf (mail_file, "Received: from %s ([%s]) by %s\n"
		                    "          (with POP3) id AA%s for <%s>;\n"
		                    "          ",
		         argv[1],
		         inet_ntoa (sock.sin_addr.s_addr),
		         Hostname,
		         prefix,
		         to);
		fputs (ptime (&time_now), mail_file);
		fprintf (mail_file, "X-POP3-UIDL: %s@%s\n", uidl, argv[1]);

		result = retrieve_mail_to_file (s, buf, LINE_BUF_LEN, mail_file);

		if (ferror (mail_file))
			disk_error = TRUE;

		if (fclose (mail_file) == EOF)
			disk_error = TRUE;

		if (result != 0 || disk_error)
		{
unlink_txt:
			unlink (mail_filename);
			sprintf (mail_filename, "%s/%s.wrk", Mailqdir, prefix);
unlink_wrk:
			unlink (mail_filename);
			rmlock (Mailqdir, prefix);
			goto send_quit;
		}

		pwait (NULL);	/*  Allow any waiting processes to go first  */
		smartdrive_commit_cache ();

		tprintf ("POP3: msg %lu To: %s From: %s (%lu)\n", message_idx, to, from, size);

		sprintf (buf, "DELE %lu\n", message_idx);
		send_pop3_command (s, buf);
		result = get_pop3_response (s, buf, LINE_BUF_LEN);
		if (result != 0)
			goto send_quit;

		rmlock (Mailqdir, prefix);

		free (to);
		to = NULL;
		free (from);
		from = NULL;
		free (uidl);
		uidl = NULL;
	}

	tputc ('\n');
	success = TRUE;

send_quit:
	send_pop3_command (s, "QUIT\n");
	if (success)
		get_pop3_response (s, buf, LINE_BUF_LEN);	/*  Wait for response  */

abort_connection:
	if (s > 0)
		close_s (s);
#ifdef	RUN_IN_SESSION
	sp->s = -1;
#endif
	
	free (to);
	free (from);
	free (uidl);
	free (buf);
	
#ifdef	RUN_IN_SESSION
	keywait (NULL, 1);
	freesession (sp);
#endif

#ifdef	notdef
	/*  Start SMTP server after successful listing  */
	if (success && argc > 4 && stricmp (argv[4], "smtp") == 0)
	{
		char	buf[16];
		int		save_input, save_output;

		/*
		 *  Substitute input, output sockets so it appears server was
		 *  started from console command session (else smtp output is
		 *  lost when this session terminates).
		 */
		save_input  = Curproc->input;
		save_output = Curproc->output;
		Curproc->input  = Cmdpp->input;
		Curproc->output = Cmdpp->output;

		/*  In a buf as cmdparse mangles the buffer  */
		strcpy (buf, "start smtp\n");
		cmdparse (Cmds, buf, NULL);

		Curproc->input  = save_input;
		Curproc->output = save_output;
	}
#endif

	return (0);
}
