/* textwin.c */

/*
   Anthony McCarthy -- August 1989
   
   No claims are made regarding efficiency, elegance, functionality,
   usability, robustness, readability, portability, etc etc :-)
*/

/*
 *	$Id: textwin.c 1.2 1997/06/20 11:26:42 Graham Exp $
 *	Jun 97	1.2	GeK	Added ip.h to fix warning
 *	      	   	   	Inline asm constants from nnnh to C style 0xnnn for cpp
 */

#include <stdlib.h>
#include <stdio.h>
#include <conio.h>
#include <dir.h>
#include <dos.h>
#include <io.h>
#include <sys/stat.h>
#include <string.h>
#include <process.h>
#include <fcntl.h>
#include <alloc.h>
#include <stdarg.h>
#include <bios.h>
#include "config.h"
#include "global.h"
#include "files.h"
#include "mbuf.h"
#include "proc.h"
#include "ip.h"			/*  GeK - fix warning  */
#include "iface.h"
#include "internet.h"
#include "session.h"
#include "socket.h"
#include "cmdparse.h"
#include "textwin.h"
#include "textwinp.h"


# ifdef TEXTWIN


/*--------------------------------------------------------------------------*/




/* Some prototypes */

static void AddWindowToTopOfList(HWINDOW pWindow);
static void RemoveWindowFromList(HWINDOW pWindow);
static void UpdateWindowsAboveAndIncluding(HWINDOW pWindow);

static USHORT FindUniqueWindowId(void);



static void PutDisplayCells( PBYTE buff,
                      int   qty,
                      int   y,
                      int   x )
{
 unsigned long p = (unsigned long)screenBuffer;

 if (y >= 0  &&  y < screenHeight)
    {
     /*
        Clip to the screen...
     */

     qty *= 2;

     if (x < 0)
        {
         qty  +=  2*x; /* x is negative */
         buff += -2*x;
         x = 0;
        }

     if (qty+2*x >= 2*screenWidth)
        qty = 2 * (screenWidth - x);

     if (qty > 0)
        _fmemcpy( (PVOID)(p | (unsigned long)(2*(y*screenWidth + x))), 
                  buff, 
                  qty
                );
    }
}


void SetCursorPos(int y, int x)
{
 BEGIN_ASM
   mov ax,0x0200
   mov bx,0x0000
   mov dh,byte ptr y
   mov dl,byte ptr x

   int 0x10
 END_ASM
}



void    SetWindowUserData(HWINDOW hWindow, PVOID userData)
{
 if (IsWindow(hWindow))
    WINUSERDATA(hWindow) = userData;
}


PVOID   GetWindowUserData(HWINDOW hWindow)
{ 
 PVOID userData = NULL;

 if (IsWindow(hWindow))
    userData = WINUSERDATA(hWindow);

 return userData;
}

static void InitVisibilityStruct( PVISIBILITY vis, 
                                  SHORT x, 
                                  SHORT y, 
                                  SHORT w, 
                                  SHORT h, 
                                  PVISBUFFER buf, 
                                  PVISFLAGS  flg)
{
 UNUSED_ARG(y);

 VISALIGN(*vis)    = x & 7;
 VISBITSWIDE(*vis) = w;
 VISWIDTH(*vis)    = CALCVISWIDTH(w);
 VISHEIGHT(*vis)   = h;
 VISBUFFER(*vis)   = buf;
 VISFLAGS(*vis)    = flg;
} /*InitVisibilityStruct*/

static void SetCompletelyVisible(PVISIBILITY v)
{
 int   yCtr;
 PBYTE p;

 for (yCtr=VISHEIGHT(*v)-1; yCtr>=0; yCtr--)
     {
      p = VISLINE(*v, yCtr);
     
      memset(p, '\xff', VISWIDTH(*v));
     
      VISLINEFLAGS(*v, yCtr) = VFL_NORMAL | VFL_WHOLELINEVISIBLE;
     }
} /*SetCompletelyVisible*/


static void SetNonVisibleRect( PVISIBILITY v, 
                               SHORT x, 
                               SHORT y, 
                               SHORT w, 
                               SHORT h )
{
 int  xCtr,
      yCtr;
 BOOL fullWidth;

 if (x < 0)
    {
     w += x; /* x is negative */
     x  = 0;
    }

 if (y < 0)
    {
     h += y; /* y is negative */
     y  = 0;
    }

 if (x+w > VISBITSWIDE(*v))
    w = VISBITSWIDE(*v) - x;

 if (y+h > VISHEIGHT(*v))
    h = VISHEIGHT(*v) - y;

 fullWidth =     (x-VISALIGN(*v)   <= 0)
             &&  (x-VISALIGN(*v)+w >= VISBITSWIDE(*v));

 for ( yCtr = y; 
       yCtr < y+h; 
       yCtr++ )
     {
      for ( xCtr = x; 
            xCtr < x+w; 
            xCtr++ )
          VISBYTE(*v, xCtr, yCtr) &= NOTBYTEBIT(VISX(*v, xCtr) & 7);
     
      VISLINEFLAGS(*v, yCtr) &= ~VFL_WHOLELINEVISIBLE;
     
      if (fullWidth)
         VISLINEFLAGS(*v, yCtr) |=  VFL_WHOLELINEHIDDEN;
     }

} /*SetNonVisibleRect*/


static void WindowCalcVisibility( PVISIBILITY screenVis, 
                                  HWINDOW     pWindow    )
{
 int   xCtr,
       yCtr;
 PBYTE src,
       dst;
 BOOL  allHidden,
       allVisible;
 int xScreenLo,
     xScreenHi,
     yScreen,
     xWinLo,
     xWinHi;
 int qty;

 if (IsWindow(pWindow))
    {
     if (WINFLAGS(pWindow) & WFL_HIDDEN)
        {
         SetNonVisibleRect( &WINVIS(pWindow),
                            0,
                            0,
                            WINWIDTH(pWindow),
                            WINHEIGHT(pWindow) );
        }
     else
        {
         for ( yCtr = 0; 
               yCtr < WINVISHEIGHT(pWindow); 
               yCtr++ )
             {
              // Grab the visibility bits for this line

              CONSISTENCY_CHECK(pWindow);
              
              yScreen   = WINYPOS(pWindow) + yCtr;

              if (yScreen >= 0  &&  yScreen < screenHeight)
                 {
                  xScreenLo = WINLEFT(pWindow);
                  xScreenHi = WINRIGHT(pWindow);

                  xWinLo    = 0;
                  xWinHi    = WINWIDTH(pWindow) - 1;

                  if (xScreenLo < 0)
                     {
                      SetNonVisibleRect( &WINVIS(pWindow), 
                                          0, 
                                          yCtr, 
                                          -xScreenLo, 
                                          1 );

                      xWinLo    = -xScreenLo;
                      xScreenLo = 0;
                     }

                  if (xScreenHi >= screenWidth)
                     {
                      SetNonVisibleRect( &WINVIS(pWindow), 
                                          screenWidth - WINLEFT(pWindow), 
                                          yCtr, 
                                          xScreenHi - screenWidth + 1, 
                                          1 );
                      

                      xWinHi    = screenWidth - 1 - WINLEFT(pWindow);
                      xScreenHi = screenWidth - 1;
                     }


                  if (xWinHi >= xWinLo)
                     {
                      qty =   WINVISBYTEOFFSET(pWindow, xWinHi)
                            - WINVISBYTEOFFSET(pWindow, xWinLo)
                            + 1;
                      
                      if (qty > 0)
                         {
                          src = (PBYTE)&VISBYTE( *screenVis,
                                                 xScreenLo,
                                                 yScreen );
                          
                          dst = (PBYTE)&WINVISBYTE(pWindow, xWinLo, yCtr);
                      
                          memcpy(dst, src, qty);

                          CONSISTENCY_CHECK(pWindow);
                         }
                     }
                 }
              else
                 SetNonVisibleRect(&WINVIS(pWindow), 0, yCtr, WINWIDTH(pWindow), 1);

              CONSISTENCY_CHECK(pWindow);

              // Determine if this line has any special visibility
              // attributes ... ie completely visible or completely
              // hidden. Partially obscured lines are relatively
              // slow to physically update, so we provide a means
              // to optimise the process ...
             
              allVisible = TRUE;
              allHidden  = TRUE;
              for ( xCtr = 0;
                    xCtr < WINWIDTH(pWindow) && (allVisible || allHidden);
                    xCtr++ )
                  if (WINCELLISVISIBLE(pWindow, xCtr, yCtr))
                     allHidden  = FALSE;
                  else
                     allVisible = FALSE;
             
              if (allVisible)
                 WINVISLINEFLAGS(pWindow, yCtr) = VFL_WHOLELINEVISIBLE;
              else if (allHidden)
                 WINVISLINEFLAGS(pWindow, yCtr) = VFL_WHOLELINEHIDDEN;
              else
                 WINVISLINEFLAGS(pWindow, yCtr) = VFL_NORMAL;
             }

         SetNonVisibleRect( screenVis,
                            WINXPOS(pWindow),
                            WINYPOS(pWindow),
                            WINWIDTH(pWindow),
                            WINHEIGHT(pWindow) );
        }
    }
} /*WindowCalcVisibility*/


static void CalcVisibilities(void)
{
 SHORT w = GetScreenWidth();
 SHORT h = GetScreenHeight();
 SHORT bytes = sizeof(VISIBILITYSTRUCT)
                + CALCVISBYTES(w, h)
                + CALCVISFLAGS(w, h);
 PVISIBILITY screenVis;


 screenVis = (PVISIBILITY)MEMALLOC(bytes);

 if (screenVis != NULL)
    {
     PVISBUFFER  visbuf;
     PVISFLAGS   visflg;
     HWINDOW     hWindow;

     visbuf = (PVISBUFFER)((PBYTE)screenVis + sizeof(VISIBILITYSTRUCT));
     visflg = (PVISFLAGS )((PBYTE)visbuf + CALCVISBYTES(w, h));

     InitVisibilityStruct( screenVis,
                           0,
                           0,
                           w,
                           h,
                           visbuf,
                           visflg );

     SetCompletelyVisible(screenVis);

     hWindow = topwindow;
     while (IsWindow(hWindow))
           {
            WindowCalcVisibility(screenVis, hWindow);
           
            hWindow = WINNEXTWINDOW(hWindow);
           }

     MEMFREE(screenVis);
    }
} /*CalcAllVisibilities*/


static void SetWindowFrameChars(HWINDOW pWindow, PSZ s)
{
  if (IsWindow(pWindow))
     {
      USHORT len = strlen(s);

#define SETCH(dst, inx) if((inx)<len) if(s[(inx)] != '\xFF') (dst)=s[(inx)]

      SETCH(WINFRAMECHV(pWindow) ,  0);
      SETCH(WINFRAMECHH(pWindow) ,  1);
      SETCH(WINFRAMECHTL(pWindow),  2);
      SETCH(WINFRAMECHBL(pWindow),  3);
      SETCH(WINFRAMECHTR(pWindow),  4);
      SETCH(WINFRAMECHBR(pWindow),  5);
      SETCH(WINFRAMECHX(pWindow) ,  6);
      SETCH(WINFRAMECHJL(pWindow),  7);
      SETCH(WINFRAMECHJR(pWindow),  8);
      SETCH(WINFRAMECHJU(pWindow),  9);
      SETCH(WINFRAMECHJD(pWindow), 10);

      WINFRAMECHARS(pWindow).resCh     = s[11];
      WINFRAMECHARS(pWindow).maxCh     = s[12];
      WINFRAMECHARS(pWindow).minCh     = s[13];
      WINFRAMECHARS(pWindow).lscrCh    = s[14];
      WINFRAMECHARS(pWindow).rscrCh    = s[15];
      WINFRAMECHARS(pWindow).uscrCh    = s[16];
      WINFRAMECHARS(pWindow).dscrCh    = s[17];
      WINFRAMECHARS(pWindow).hthumbCh  = s[18];
      WINFRAMECHARS(pWindow).vthumbCh  = s[19];
      WINFRAMECHARS(pWindow).hscrbarCh = s[20];
      WINFRAMECHARS(pWindow).vscrbarCh = s[21];

#undef SETCH

     }
} /*SetWindowFrameChars*/


HWINDOW GetDesktopWindow(void)
{
 HWINDOW hDesktop = NULL;

 if (IsWindow(desktopwindow))
    hDesktop = desktopwindow;

 return hDesktop;
} /*GetDesktopWindow*/


HWINDOW GetWindowTop(void)
{
 HWINDOW hTop = NULL;

 if (IsWindow(topwindow))
    hTop = topwindow;

 return hTop;
} /*GetWindowTop*/


HWINDOW GetWindowFocus(void)
{
 HWINDOW pFocus = NULL;

 if (IsWindow(focuswindow))
    pFocus = focuswindow;

 return pFocus;
} /*GetWindowFocus*/



BOOLEAN SetWindowFocus(HWINDOW newFocus)
{
 BOOLEAN ok = FALSE;

 if (IsWindow(newFocus))
    {
     BringWindowToTop(newFocus);

     if (newFocus != GetWindowFocus())
        {
         if (IsWindow(focuswindow))
            {
             SetWindowFrameChars(focuswindow, singleFrameChars);
             FrameWindow(focuswindow);
            }

         focuswindow = newFocus;

         SetWindowFrameChars(focuswindow, doubleFrameChars);
         FrameWindow(focuswindow);

         UpdateCursorDisplay(focuswindow);
        }

     ok = TRUE;
    }

 return ok;
} /*SetWindowFocus*/


static void CalcClientArea(HWINDOW pWindow)
{
 SHORT borderX,
       borderY;

 if (IsWindow(pWindow))
    {
     borderX = (WINFLAGS(pWindow) & WFL_NOVERTBORDER) ? 0 : 1;
     borderY = (WINFLAGS(pWindow) & WFL_NOHORZBORDER) ? 0 : 1;

     WINCLIENTXPOS(pWindow)   = WINXPOS(pWindow)   + borderX;
     WINCLIENTYPOS(pWindow)   = WINYPOS(pWindow)   + borderY;
     WINCLIENTWIDTH(pWindow)  = WINWIDTH(pWindow)  - borderX * 2;
     WINCLIENTHEIGHT(pWindow) = WINHEIGHT(pWindow) - borderY * 2;
    }
} /*CalcClientArea*/



#if 0
static int ClientXToWindowX(HWINDOW h, int x)
{ 
 if (WINFLAGS(h) & WFL_NOVERTBORDER)
    return x;
 else
    return x + 1;
}
#endif


#if 0
static int ClientYToWindowY(HWINDOW h, int y)
{ 
 if (WINFLAGS(h) & WFL_NOHORZBORDER)
    return y;
 else
    return y + 1;
}
#endif



HWINDOW PrimAllocWindow(USHORT x, USHORT y, USHORT w, USHORT h, int logWidth, int logHeight)
{
 PW     pWindow = NULL;
 PBYTE  pTmpWinBuffer = NULL,
        pTmpVisbuf    = NULL,
        pTmpVisflg    = NULL,
        pTmpBorder    = NULL;
 PBYTE  visbuf;
 PBYTE  visflg;


 TextwinInit();


 pTmpWinBuffer = MEMALLOC(2 * logWidth * logHeight);
 pTmpBorder    = MEMALLOC(2 * max(w,h));
 pTmpVisbuf    = MEMALLOC(CALCVISBYTES(w, h));
 pTmpVisflg    = MEMALLOC(CALCVISFLAGS(w, h));
 pWindow       = (PW)MEMALLOC(sizeof(*pWindow));

 if (! (pWindow  &&  pTmpWinBuffer  &&  pTmpVisbuf  &&  pTmpVisflg))
    {
     if (pTmpWinBuffer) MEMFREE(pTmpWinBuffer);
     if (pTmpBorder   ) MEMFREE(pTmpBorder   );
     if (pTmpVisbuf   ) MEMFREE(pTmpVisbuf   );
     if (pTmpVisflg   ) MEMFREE(pTmpVisflg   );
     if (pWindow      ) MEMFREE(pWindow      );
    
     pWindow = NULL;
    }
 else
    {
     memset(pWindow, 0, sizeof(*pWindow));

     WINBORDERBUFFER(pWindow)  = pTmpBorder;
     WINLOGICALBUFFER(pWindow) = pTmpWinBuffer;
     WINLOGICALX(pWindow) = 0;
     WINLOGICALY(pWindow) = 0;
     WINLOGICALWIDTH(pWindow)  = logWidth;
     WINLOGICALHEIGHT(pWindow) = logHeight;

     visbuf = pTmpVisbuf;
     visflg = pTmpVisflg;

     WINXPOS(pWindow)   = 0; // NOTE: Window pos is set 'properly' below
     WINYPOS(pWindow)   = 0;
     WINWIDTH(pWindow)  = w;
     WINHEIGHT(pWindow) = h;

     InitVisibilityStruct( &WINVIS(pWindow),
                           WINXPOS(pWindow),
                           WINYPOS(pWindow),
                           WINWIDTH(pWindow),
                           WINHEIGHT(pWindow),
                           visbuf,
                           visflg );

     // Next two for debugging only ...
     WINVISLINEFLAGS(pWindow, 0) = (BYTE)0;
     WINVISLINEFLAGS(pWindow, WINHEIGHT(pWindow)-1) = (BYTE)0;

     WINDEFATTR(pWindow)        = defaultWindowAttribute;
     WINBORDERATTR(pWindow)     = defaultWindowAttribute;
     WINBACKGROUNDCHAR(pWindow) = 0x20;


     // NOTE: Created invisible !

     WINFLAGS(pWindow) = WFL_NORMAL | WFL_HIDDEN;

     if (! wantBorders)
        WINFLAGS(pWindow) |= WFL_NOBORDER;


     // This statement marks the window structure as being
     // valid. Can now use pWindow as parameter to other
     // routines.

     WINMAGIC(pWindow) = WINDOWMAGIC;

     ResetDirtyArea(pWindow);

     WINID(pWindow) = FindUniqueWindowId();

     SetWindowFrameChars(pWindow, singleFrameChars);

     CalcClientArea(pWindow);

     AddWindowToTopOfList(pWindow);

     WindowSetPos(pWindow, x, y);

     ClearWindow(pWindow);

     ShowWindow(pWindow, TRUE);
    }

 return (HWINDOW)pWindow;
} /*PrimAllocWindow*/


HWINDOW AllocWindow(USHORT x, USHORT y, USHORT w, USHORT h)
{
 int logX,
     logY;
 int borderDX;
 int borderDY;

 TextwinInit();

 borderDX = wantBorders ? 2 : 0;
 borderDY = wantBorders ? 2 : 0;

 logX = defaultLogicalWidth; /* AMcC : 29Jul92 : changed ',' to ';'! */
 logY = defaultLogicalHeight;

 w = max(5, min(w, logX + borderDX));
 h = max(2, min(h, logY + borderDY));

 return PrimAllocWindow(x,y,w,h,logX,logY);
}


HWINDOW AllocTerminalWindow(USHORT x, USHORT y, USHORT w, USHORT h)
{
 int borderDX;
 int borderDY;

 TextwinInit();

 borderDX = wantBorders ? 2 : 0;
 borderDY = wantBorders ? 2 : 0;

 w = max(1, w);
 h = max(1, h);

 return PrimAllocWindow(x,y,w+borderDX,h+borderDY,w,h);
}


HWINDOW DefaultAllocWindow(void)
{
 int x,
     y;

 TextwinInit();

 x = defaultWindowX;
 y = defaultWindowY;

 if (cascadeOffsetX >= cascadeRangeX)
    cascadeOffsetX = 0;

 if (cascadeOffsetY >= cascadeRangeY)
    cascadeOffsetY = 0;

 x = defaultWindowX + cascadeOffsetX;
 y = defaultWindowY + cascadeOffsetY;

 cascadeOffsetX++;
 cascadeOffsetY++;

 return AllocWindow( x,
                     y,
                     defaultWindowDX,
                     defaultWindowDY );
}


void FreeWindow(HWINDOW pWindow)
{
 HWINDOW next;

 if (IsWindow(pWindow))
    {
     next = WINNEXTWINDOW(pWindow);

     if (pWindow == focuswindow)
       SetWindowFocus(IsWindow(next) ? next : desktopwindow);

     RemoveWindowFromList(pWindow);

     if (WINCAPTION(pWindow) != NULL)
        {
         MEMFREE(WINCAPTION(pWindow));
         WINCAPTION(pWindow) = NULL;
        }

     MEMFREE(WINBORDERBUFFER(pWindow));
     MEMFREE(WINLOGICALBUFFER(pWindow));
     MEMFREE(WINVISBUFFER(pWindow));
     MEMFREE(WINVISFLAGS(pWindow));
     MEMFREE(pWindow);

     CalcVisibilities();

     UpdateWindowsBelowAndIncluding(next);
    }
} /*FreeWindow*/


void ShowWindowNoUpdate(HWINDOW hWindow, BOOL show)
{
 if (IsWindow(hWindow))
    {
     if (show)
       WINFLAGS(hWindow) &= ~WFL_HIDDEN;
     else
       WINFLAGS(hWindow) |=  WFL_HIDDEN;

     CalcVisibilities();
    }
} /*ShowWindow*/


void ShowWindow(HWINDOW hWindow, BOOL show)
{
 if (IsWindow(hWindow))
    {
     ShowWindowNoUpdate(hWindow, show);

     UpdateWindowsBelowAndIncluding(hWindow);

     if (show)
        if (hWindow == GetWindowFocus())
           UpdateCursorDisplay(hWindow);
    }
} /*ShowWindow*/


void WindowSetPosNoUpdate(HWINDOW hWindow, SHORT x, SHORT y)
{
  if (IsWindow(hWindow))
     {
      WINXPOS(hWindow) = x;
      WINYPOS(hWindow) = y;

      InitVisibilityStruct( &WINVIS(hWindow),
                            WINXPOS(hWindow),
                            WINYPOS(hWindow),
                            WINWIDTH(hWindow),
                            WINHEIGHT(hWindow),
                            WINVISBUFFER(hWindow),
                            WINVISFLAGS(hWindow)
                          );

      CalcClientArea(hWindow);

      CalcVisibilities();
     }
} /*WindowSetPosNoUpdate*/


void EnsureCursorIsInWindow(HWINDOW pWindow)
{
 int curY  = WINCURY(pWindow),
     logY  = WINLOGICALY(pWindow),
     logX  = WINLOGICALX(pWindow),
     logDy = WINLOGICALHEIGHT(pWindow),
     logDx = WINLOGICALWIDTH(pWindow),
     cliDy = WINCLIENTHEIGHT(pWindow),
     cliDx = WINCLIENTWIDTH(pWindow),
     newlogY,
     newlogX;
 BOOL bUpdate = FALSE;

 if (     curY >= 0  
      &&  curY <  logDy )
    {
     newlogX = logX;
     newlogY = logY;

     if (logY + cliDy >= logDy) /* resized too large vertically? */
        {
         newlogY = max(0, logDy - cliDy);
         bUpdate = TRUE;
        }

# if 0

     /*
        This code makes the cursor appear in the window.
        I don't like this anymore - now that scrollbars are available.
     */

     if (     curY < logY
          ||  curY >= logY + cliDy)
        {
         newlogY = max(0, curY - cliDy + 1);
         bUpdate = TRUE;
        }
# endif


     if (logX + cliDx >= logDx) /* resized too large horizontally? */
        {
         newlogX = max(0, logDx - cliDx);
         bUpdate = TRUE;
        }

     if (bUpdate)
        {
         WINLOGICALX(pWindow) = newlogX;
         WINLOGICALY(pWindow) = newlogY;
     
         UpdateWindow(pWindow);
     
         if (GetWindowFocus() == pWindow)
            UpdateCursorDisplay(pWindow);
        }
    }
}


void ResizeWindow(HWINDOW hWindow, SHORT dx, SHORT dy)
{
 PW    pWindow    = (PW)hWindow;
 PBYTE pTmpVisbuf,
       pTmpVisflg,
       pTmpBorder,
       pOldVisbuf,
       pOldVisflg,
       pOldBorder;
 int   oldx, 
       oldy;
 int   borderX,
       borderY;


 borderX = (WINFLAGS(pWindow) & WFL_NOVERTBORDER) ? 0 : 1;
 borderY = (WINFLAGS(pWindow) & WFL_NOHORZBORDER) ? 0 : 1;

 dx = max(0, min(dx, 2*borderX + WINLOGICALWIDTH(pWindow )));
 dy = max(0, min(dy, 2*borderY + WINLOGICALHEIGHT(pWindow)));

 if (     dx != WINWIDTH(pWindow)
      ||  dy != WINHEIGHT(pWindow) )
    {
     pTmpVisbuf = MEMALLOC(CALCVISBYTES(dx, dy));
     pTmpVisflg = MEMALLOC(CALCVISFLAGS(dx, dy));
     pTmpBorder = MEMALLOC(max(dx,dy) * 2);

     if (    pTmpVisbuf == NULL
          || pTmpVisflg == NULL
          || pTmpBorder == NULL )
        {
         if (pTmpVisbuf != NULL) 
            MEMFREE(pTmpVisbuf);
     
         if (pTmpVisflg != NULL) 
            MEMFREE(pTmpVisflg);
     
         if (pTmpBorder != NULL) 
            MEMFREE(pTmpBorder);
        }
     else
     if (IsWindow(hWindow))
        {
         ShowWindowNoUpdate(pWindow, FALSE);
     
         RemoveWindowFromList(pWindow);
     
         pOldVisbuf = WINVISBUFFER(pWindow);
         pOldVisflg = WINVISFLAGS(pWindow);
         pOldBorder = WINBORDERBUFFER(pWindow);
     
         oldx  = WINLEFT(pWindow);
         oldy  = WINTOP(pWindow);
     
         WINWIDTH(pWindow)  = dx;
         WINHEIGHT(pWindow) = dy;
     
         WINBORDERBUFFER(pWindow) = pTmpBorder;

         InitVisibilityStruct( &WINVIS(pWindow),
                                WINXPOS(pWindow),
                                WINYPOS(pWindow),
                                WINWIDTH(pWindow),
                                WINHEIGHT(pWindow),
                                pTmpVisbuf,
                                pTmpVisflg  );
     
         CalcClientArea(hWindow);
     
         MEMFREE(pOldVisbuf);
         MEMFREE(pOldVisflg);
         MEMFREE(pOldBorder);
     
         AddWindowToTopOfList(pWindow);
         WindowSetPosNoUpdate(hWindow, oldx, oldy);
     
         EnsureCursorIsInWindow(pWindow);   
     
         ShowWindow(pWindow, TRUE);
     
         UpdateWindowsBelowAndIncluding(hWindow);
        }
    }
}


void SetWindowID(HWINDOW pWindow, USHORT id)
{
 if (IsWindow(pWindow))
    WINID(pWindow) = id;
} /*SetWindowID*/


USHORT GetWindowID(HWINDOW pWindow)
{
 USHORT id = 0;

 if (IsWindow(pWindow))
    id = WINID(pWindow);

 return id;
} /*GetWindowID*/


HWINDOW FindWindowFromId(USHORT id)
{
 HWINDOW h = NULL;
 int timeout = 1000; /* In case of stupid loops */

 h = GetWindowTop();

 while (h != NULL  &&  timeout--)
       {
        if (GetWindowID(h) == id)
           return h;

        h = WINNEXTWINDOW(h);
       }

 return NULL;
}


static USHORT FindUniqueWindowId(void)
{
 USHORT id = 2; /* Desktop window has id 1, always */

 while (FindWindowFromId(id) != NULL)
       id++;

 return id;
}


static int CalcScrollThumbPos( int sBeg,
                               int sEnd,
                               int sMin,
                               int sMax,
                               int sVal )
{
 int  sPos;
 long cells;

 sVal = max(sMin, min(sVal, sMax));

 if (sVal <= sMin)
    sPos = sBeg;
 else if (sVal >= sMax)
    sPos = sEnd;
 else  
    {
     cells = sEnd - sBeg + 1;

     sPos = sBeg + (int)((cells * (sVal - sMin)) / (sMax - sMin + 1));

     sPos = max(sBeg, min(sPos, sEnd));
    }

 return sPos;
}


/*
   EGODS. This routine is a bomb site...
*/

void FrameWindow(HWINDOW pWindow)
{
 int pos;
 int winX,
     winY;

 int winH,
     winW;

 BYTE  vCh, 
       hCh, 
       tlCh,
       blCh,
       trCh,
       brCh,
       resCh,
       maxCh,
       minCh,
       lscrCh,
       rscrCh,
       uscrCh,
       dscrCh,
       hthumbCh,
       vthumbCh,
       hscrbarCh,
       vscrbarCh;

 BYTE cell[2];
 int scrX,
     scrY;
 PSZ  caption;
 PBYTE pLine,
       ptrLine;


 if (IsWindow(pWindow))
    if (! (WINFLAGS(pWindow) & WFL_NOBORDER))
       {
        winH = WINHEIGHT(pWindow);
        winW = WINWIDTH( pWindow);

        HideMouse();

        vCh  = WINFRAMECHV(pWindow);
        hCh  = WINFRAMECHH(pWindow);
        tlCh = WINFRAMECHTL(pWindow);
        blCh = WINFRAMECHBL(pWindow);
        trCh = WINFRAMECHTR(pWindow);
        brCh = WINFRAMECHBR(pWindow);

        resCh     = WINFRAMECHARS(pWindow).resCh   ;
        maxCh     = WINFRAMECHARS(pWindow).maxCh   ;
        minCh     = WINFRAMECHARS(pWindow).minCh   ;
        lscrCh    = WINFRAMECHARS(pWindow).lscrCh  ;
        rscrCh    = WINFRAMECHARS(pWindow).rscrCh  ;
        uscrCh    = WINFRAMECHARS(pWindow).uscrCh  ;
        dscrCh    = WINFRAMECHARS(pWindow).dscrCh  ;
        hthumbCh  = WINFRAMECHARS(pWindow).hthumbCh;
        vthumbCh  = WINFRAMECHARS(pWindow).vthumbCh;
        hscrbarCh = WINFRAMECHARS(pWindow).hscrbarCh;
        vscrbarCh = WINFRAMECHARS(pWindow).vscrbarCh;
            
        pLine = ptrLine = WINBORDERBUFFER(pWindow);

        if (! WINWHOLELINEISHIDDEN(pWindow, 0))
           {
            memset(pLine, WINBORDERATTR(pWindow), 2*winW);
            
            
            /* THE TOP BORDER */
            
            *pLine = tlCh;
            pLine += 2;
            
            *pLine = hCh;
            pLine += 2;
            
            winX = 2;
            
            caption = WINCAPTION(pWindow);
            
            if (caption != NULL  &&  *caption)
               {
                if (winX < winW - 5)
                   {
                    *pLine = ' ';
                    pLine += 2;
            
                    winX++;
                   }
            
                for ( ;
                      winX < winW - 4;
                      winX++ )
                    {
                     if (*caption == '\0')
                        break;
                       
                     *pLine = *caption++;
                     pLine += 2;
                    }
            
                if (winX < winW - 5)
                   {
                    *pLine = ' ';
                    pLine += 2;
            
                    winX++;
                   }
               }
            
            for ( ;
                  winX < winW - 3;
                  winX++ )
                {
                 *pLine = hCh;
                 pLine += 2;
                }
            
            *pLine = (WINFLAGS(pWindow) & WFL_MINIMIZED) ? resCh : minCh;
            pLine += 2;

            *pLine = (WINFLAGS(pWindow) & WFL_MAXIMIZED) ? resCh : maxCh;
            pLine += 2;
            
            *pLine = trCh;
            
            if (WINWHOLELINEISVISIBLE(pWindow, 0))
               PutDisplayCells( ptrLine, 
                                winW, 
                                WINTOP(pWindow), 
                                WINLEFT(pWindow) );
            else
               for ( winX = 0;
                     winX < winW;
                     winX++ )
                   if (WINCELLISVISIBLE(pWindow, winX, 0))
                      PutDisplayCells( ptrLine + 2*winX, 
                                       1, 
                                       WINTOP(pWindow), 
                                       WINLEFT(pWindow) + winX);
                      
           }


        /* THE BOTTOM BORDER */

        winY = winH - 1;
        scrY = WINBOTTOM(pWindow);

        if (! WINWHOLELINEISHIDDEN(pWindow, winY))
           {
            pLine = ptrLine;
            memset(pLine, WINBORDERATTR(pWindow), 2*winW);
            
            *pLine = blCh;
            pLine += 2;
            
            for ( winX = 1;
                  winX < winW - 1;
                  winX++ )
                {
                 *pLine = hscrbarCh;
                 pLine += 2;
                }
            
            *pLine = brCh;
            

            if (winW >= 4)
               {
                ptrLine[2 * (1     )] = lscrCh;
                ptrLine[2 * (winW-2)] = rscrCh;

                if (winW >= 7)
                   { 
                    pos = CalcScrollThumbPos( 2,
                                              winW - 3,
                                              0,
                                              WINLOGICALWIDTH(pWindow) 
                                                - WINCLIENTWIDTH(pWindow), 
                                              WINLOGICALX(pWindow) );
            
                    ptrLine[2 * pos] = hthumbCh;
                   }

               }


            if (WINWHOLELINEISVISIBLE(pWindow, winY))
               PutDisplayCells( ptrLine, 
                                winW, 
                                scrY, 
                                WINLEFT(pWindow) );
            else
               for ( winX = 0;
                     winX < winW;
                     winX++ )
                   if (WINCELLISVISIBLE(pWindow, winX, winY))
                      PutDisplayCells( ptrLine + 2*winX, 
                                       1, 
                                       scrY, 
                                       WINLEFT(pWindow) + winX);
           }


        /* THE LEFT BORDER */

        cell[0] = vCh;
        cell[1] = (BYTE)WINBORDERATTR(pWindow);


        winX = 0;
        scrX = WINLEFT(pWindow);

        winY = 0;
        scrY = WINTOP(pWindow) + 1;

        if (scrX >= 0  &&  scrX < screenWidth)
           for ( winY = 1;
                 winY < winH - 1;
                 winY++, scrY++ )
               if (scrY >= 0  &&  scrY < screenWidth)
                  if (WINCELLISVISIBLE(pWindow, winX, winY))
                     PutDisplayCells(cell, 1, scrY, scrX);


        /* The ALL NEW Right Border */

        winX = winW - 1;
        scrX = WINRIGHT(pWindow);

        winY = 0;
        scrY = WINTOP(pWindow) + 1;

        if (scrX >= 0  &&  scrX < screenWidth)
           {
            pLine = ptrLine;
            memset(pLine, WINBORDERATTR(pWindow), 2*winH);
            
            for ( winY = 1;
                  winY < winH - 1;
                  winY++ )
                pLine[2 * winY] = vscrbarCh;
            
            if (winH >= 4)
               {
                pLine[2 * (1     )] = uscrCh;
                pLine[2 * (winH-2)] = dscrCh;
            
                if (winH >= 7)
                   { 
                    pos = CalcScrollThumbPos( 2,
                                              winH - 3,
                                              0,
                                              WINLOGICALHEIGHT(pWindow) 
                                                - WINCLIENTHEIGHT(pWindow), 
                                              WINLOGICALY(pWindow) );
            
                    pLine[2 * pos] = vthumbCh;
                   }
               }

            for ( winY = 1;
                  winY < winH - 1;
                  winY++, scrY++ )
                if (scrY >= 0  &&  scrY < screenWidth)
                   if (WINCELLISVISIBLE(pWindow, winX, winY))
                      PutDisplayCells(ptrLine + 2*winY, 1, scrY, scrX);
           }

        ShowMouse();
       }
} /*FrameWindow*/


void SetWindowCaption(HWINDOW pWindow, PSZ caption)
{
 PSZ newcaption;
 int len;

 if (IsWindow(pWindow))
    {
     if (WINCAPTION(pWindow) != NULL)
        {
         MEMFREE(WINCAPTION(pWindow));
         WINCAPTION(pWindow) = NULL;
        }

     if (caption != NULL)
        {
         len = strlen(caption);

         if (len > 0)
            {
             newcaption = MEMALLOC(len + 1);
            
             strcpy(newcaption, caption);
             WINCAPTION(pWindow) = newcaption;
            }
        }

     FrameWindow(pWindow);
    }
}


static void RemoveWindowFromList(HWINDOW pWindow)
{
 HWINDOW prev, 
         next;
 
 if (IsWindow(pWindow))
    {
     prev = WINPREVWINDOW(pWindow);
     next = WINNEXTWINDOW(pWindow);

     if (IsWindow(prev))
        WINNEXTWINDOW(prev) = next;

     if (IsWindow(next))
        WINPREVWINDOW(next) = prev;
   
     if (topwindow == pWindow)
        topwindow = next;
    }
} /*RemoveWindowFromList*/


static void AddWindowToTopOfList(HWINDOW pWindow)
{
 if (IsWindow(pWindow))
    {
     WINNEXTWINDOW(pWindow) = topwindow;
     WINPREVWINDOW(pWindow) = NULL;

     if (IsWindow(topwindow))
        WINPREVWINDOW(topwindow) = pWindow;

     topwindow = pWindow;
    }
} /*AddWindowToTopOfList*/


static void AddWindowBehindWindowInList(HWINDOW pWindow, HWINDOW addBehind)
{
 if (IsWindow(pWindow))
    if (! IsWindow(addBehind))
       AddWindowToTopOfList(pWindow);
    else
       {
        HWINDOW next = WINNEXTWINDOW(addBehind);

        if (IsWindow(next))
           WINPREVWINDOW(next) = pWindow;
        WINNEXTWINDOW(addBehind) = pWindow;

        WINPREVWINDOW(pWindow) = addBehind;
        WINNEXTWINDOW(pWindow) = next;
       }
} /*AddWindowBehindWindowInList*/



HWINDOW GetWindowBottom(void)
{
 HWINDOW bottomWindow = GetWindowTop();
 USHORT  timeout = 100;

 // Find the bottom-most window
 while (IsWindow(WINNEXTWINDOW(bottomWindow)) && --timeout)
       {
        if (WINNEXTWINDOW(bottomWindow) == desktopwindow)
           break;

        bottomWindow = WINNEXTWINDOW(bottomWindow);
       }

 return bottomWindow;
}


void UpdateWindowsBelowAndIncluding(HWINDOW pWindow)
{
 if (IsWindow(pWindow))
    {
     HWINDOW bottomWindow = pWindow;
     HWINDOW p = pWindow;
     USHORT  timeout = 100;

     // Find the bottom-most window
     while (IsWindow(WINNEXTWINDOW(bottomWindow)) && --timeout)
           bottomWindow = WINNEXTWINDOW(bottomWindow);

     // Now update all windows (from the bottom up) until we hit
     // pWindow ...

     p = bottomWindow;
     timeout = 100;
     do {
          UpdateWindow(p);
          FrameWindow(p);

          p = WINPREVWINDOW(p);
        } while (IsWindow(p)  &&  WINNEXTWINDOW(p) != pWindow  &&  --timeout);
    }
} /*UpdateWindowsBelowAndIncluding*/


static void UpdateWindowsAboveAndIncluding(HWINDOW pWindow)
{
 if (IsWindow(pWindow))
    {
     HWINDOW p = pWindow;
     USHORT  timeout = 100;

     while (IsWindow(p)  &&  --timeout)
       {
         UpdateWindow(p);
         FrameWindow(p);

         p = WINPREVWINDOW(p);
       }
    }
} /*UpdateWindowsAboveAndIncluding*/


BOOLEAN MoveWindowBehind(HWINDOW pWindow, HWINDOW pWindowBehind)
{
 BOOLEAN ok = FALSE;

 if (IsWindow(pWindow))
    if (pWindow != pWindowBehind)
       {
        RemoveWindowFromList(pWindow);
        AddWindowBehindWindowInList(pWindow, pWindowBehind);
    
        CalcVisibilities();
    
        UpdateWindowsAboveAndIncluding(pWindow);
       }

 return ok;
} /*MoveWindowBehind*/


void BringWindowToTop(HWINDOW hWindow)
{
 if (hWindow != GetWindowTop())
    MoveWindowBehind(hWindow, NULL);
}


void WindowSetPos(HWINDOW hWindow, SHORT x, SHORT y)
{
  if (IsWindow(hWindow))
     {
      WindowSetPosNoUpdate(hWindow, x, y);

      UpdateWindowsBelowAndIncluding(hWindow);

      /* AMcC : 20Jun92 : move the cursor too if we're moving
      ** the window that contains the cursor!
      */
      if (hWindow == GetWindowFocus())
         UpdateCursorDisplay(hWindow);
     }
} /*WindowSetPos*/


static BOOLEAN WindowContains(HWINDOW pWindow, SHORT x, SHORT y)
{
 return (     x >= WINLEFT(pWindow)
          &&  x <= WINRIGHT(pWindow)
          &&  y >= WINTOP(pWindow)
          &&  y <= WINBOTTOM(pWindow) );
} /*WindowContains*/


#if 0
static BOOLEAN IsScreenXYVisibleAboveWindow(HWINDOW pWindow, SHORT x, SHORT y)
{
  BOOLEAN isVisible = FALSE;

  if (IsWindow(pWindow))
    {
      HWINDOW p = WINPREVWINDOW(pWindow);
      SHORT   timeout = 100;

      isVisible = TRUE;

      while (isVisible  &&  IsWindow(p)  &&  --timeout)
        {
          isVisible = ! WindowContains(p, x, y);
          p = WINPREVWINDOW(p);
        }
    }

  return isVisible;
} /*IsScreenXYVisibleAboveWindow*/
#endif


#if 0
static BOOLEAN WindowCellVisible(HWINDOW pWindow, SHORT x, SHORT y)
{
 BOOLEAN isVisible = FALSE;

 if (IsWindow(pWindow))
    {
     SHORT screenX = WINXPOS(pWindow) + x;
     SHORT screenY = WINYPOS(pWindow) + y;

     isVisible = IsScreenXYVisibleAboveWindow(pWindow, screenX, screenY);
    }

 return isVisible;
} /*WindowCellVisible*/
#endif


BOOL IsVisible(HWINDOW w)
{
 return (! WINISHIDDEN(w));
}


HWINDOW WindowFromPoint(SHORT x, SHORT y)
{
 HWINDOW w = GetWindowTop();

 while (w != NULL)
       if (    IsVisible(w)
            && WindowContains(w, x, y) )
          break;
       else          
          w = WINNEXTWINDOW(w);

 return w;
}

HitType WindowHitType(HWINDOW w, SHORT x, SHORT y)
{
 HitType ht = HT_NULL;
 int thumbPos;

 if (IsWindow(w))
    if (IsVisible(w)  &&  WindowContains(w, x, y))
       {
        ht = HT_CLIENT;

        if (y == WINTOP(w))
           {
            ht = HT_CAPTION;

            if (x == WINLEFT(w))
               ht = HT_CLOSE;
            else if (x == WINRIGHT(w)-2)
               {
                if (WINFLAGS(w) & WFL_MINIMIZED)
                   ht = HT_RESTORE;
                else
                   ht = HT_MINIMIZE;
               }
            else if (x == WINRIGHT(w)-1)
               {
                if (WINFLAGS(w) & WFL_MAXIMIZED)
                   ht = HT_RESTORE;
                else
                   ht = HT_MAXIMIZE;
               }
           }
        else if (y == WINBOTTOM(w))
           {
            if (x == WINRIGHT(w))
               ht = HT_SIZE;
            else if (x == WINLEFT(w)+1)
               ht = HT_SCROLL_LEFT;
            else if (x == WINRIGHT(w)-1)
               ht = HT_SCROLL_RIGHT;
            else 
               {
                thumbPos = CalcScrollThumbPos( WINLEFT(w) + 2,
                                               WINRIGHT(w) - 2,
                                               0,
                                               WINLOGICALWIDTH(w)
                                                - WINCLIENTWIDTH(w),
                                               WINLOGICALX(w) );

                if (x == thumbPos)
                   ht = HT_SCROLL_HORZTHUMB;
                else if (x > WINLEFT(w)+1  &&  x < thumbPos)
                   ht = HT_SCROLL_PAGELEFT;
                else if (x > thumbPos  &&  x < WINRIGHT(w)-1)
                   ht = HT_SCROLL_PAGERIGHT;
               }
           }
        else if (x == WINRIGHT(w))
           {
            if (y == WINTOP(w)+1)
               ht = HT_SCROLL_UP;
            else if (y == WINBOTTOM(w)-1)
               ht = HT_SCROLL_DOWN;
            else
               {
                thumbPos = CalcScrollThumbPos( WINTOP(w) + 2,
                                               WINBOTTOM(w) - 2,
                                               0,
                                               WINLOGICALHEIGHT(w)
                                                - WINCLIENTHEIGHT(w),
                                               WINLOGICALY(w) );

                if (y == thumbPos)
                   ht = HT_SCROLL_VERTTHUMB;
                else if (y > WINTOP(w)+1  &&  y < thumbPos)
                   ht = HT_SCROLL_PAGEUP;
                else if (y > thumbPos  &&  y < WINBOTTOM(w)-1)
                   ht = HT_SCROLL_PAGEDOWN;
               }
           }
       }

 return ht;
}


void UpdateClientRect( HWINDOW pWindow, 
                       int loX,
                       int hiX,
                       int loY, 
                       int hiY )
{
 SHORT xCtr;
 SHORT yCtr;
 SHORT xWin;
 SHORT yWin;
 int   width;
 int   qty;
 int   borderX,
       borderY;
 int   xLog,
       yLog;


 if (IsWindow(pWindow))
    {
     if (    ! WINDISABLEUPDATE(pWindow)
          && ! WINISHIDDEN(pWindow) )
        { 
         HideMouse();

         if (WINFLAGS(pWindow) & WFL_NOBORDER)
            borderX = borderY = 0;
         else
            borderX = borderY = 1;

         yLog = loY + WINLOGICALY(pWindow);

         for ( yCtr  = loY; 
               yCtr <= hiY; 
               yCtr++, yLog++ )
           if (! WINWHOLELINEISHIDDEN(pWindow, yCtr+borderY))
              {
               xWin = WINCLIENTXPOS(pWindow) + loX;
               yWin = WINCLIENTYPOS(pWindow) + yCtr;
               width = hiX - loX + 1;

               if (WINWHOLELINEISVISIBLE(pWindow, yCtr+borderY))
                  PutDisplayCells( &WINCHARXY( pWindow, 
                                               loX  + WINLOGICALX(pWindow), 
                                               yLog
                                             ),
                                   width,
                                   yWin,
                                   xWin );
               else
                  {
                   xCtr = loX;
                   while (xCtr <= hiX)
                         if (! WINCELLISVISIBLE(pWindow, xCtr+borderX, yCtr+borderY))
                            {
                             xCtr++;
                             xWin++;
                            }
                         else
                            {
                             xLog = xCtr + WINLOGICALX(pWindow);
                         
                             for ( qty = 0;
                                      xCtr <= hiX
                                   && WINCELLISVISIBLE(pWindow, xCtr+borderX, yCtr+borderY);
                                   xCtr++ )
                                 qty++;
                         
                             PutDisplayCells( &WINCHARXY(pWindow, xLog, yLog),
                                              qty,
                                              yWin,
                                              xWin );
                         
                             xWin += qty;
                            }
                  }
              }

         ShowMouse();
        }
    }
}


void UpdateLogicalRect( HWINDOW hWindow,
                        int loX,
                        int hiX,
                        int loY,
                        int hiY )
{
 int cliLoX,
     cliHiX,
     cliLoY,
     cliHiY;

 if (IsWindow(hWindow)) 
    {
     cliLoX = loX - WINLOGICALX(hWindow);
     cliHiX = hiX - WINLOGICALX(hWindow);
     cliLoY = loY - WINLOGICALY(hWindow);
     cliHiY = hiY - WINLOGICALY(hWindow);

     if (    cliHiX >= 0
          && cliLoX <  WINCLIENTWIDTH(hWindow)
          && cliHiY >= 0 
          && cliLoY <  WINCLIENTHEIGHT(hWindow) )
        {
         cliLoX = max(0, min(cliLoX, WINCLIENTWIDTH( hWindow)-1));
         cliHiX = max(0, min(cliHiX, WINCLIENTWIDTH( hWindow)-1));
         cliLoY = max(0, min(cliLoY, WINCLIENTHEIGHT(hWindow)-1));
         cliHiY = max(0, min(cliHiY, WINCLIENTHEIGHT(hWindow)-1));
         
         UpdateClientRect(hWindow, cliLoX, cliHiX, cliLoY, cliHiY);
        }
    }
}



void UpdateWindow(HWINDOW pWindow)
{
 UpdateClientRect( pWindow, 
                   0, 
                   WINCLIENTWIDTH(pWindow)-1, 
                   0, 
                   WINCLIENTHEIGHT(pWindow)-1 );
} /*UpdateWindow*/


void SetWindowAttribute(HWINDOW pWindow, BYTE attribute)
{
 if (IsWindow(pWindow))
    WINDEFATTR(pWindow) = attribute;
} /*SetWindowAttribute*/


void SetBorderAttribute(HWINDOW pWindow, BYTE attribute)
{
 if (IsWindow(pWindow))
    WINBORDERATTR(pWindow) = attribute;
} /*SetBorderAttribute*/


BYTE GetWindowAttribute(HWINDOW pWindow)
{
 BYTE attr = 0;

 if (IsWindow(pWindow))
    attr = (BYTE)WINDEFATTR(pWindow);

 return attr;
} /*GetWindowAttribute*/


void EnableWindowUpdate(HWINDOW pWindow, BOOLEAN enable)
{
 if (IsWindow(pWindow))
    if (enable)
       {
        if (WINDISABLEUPDATE(pWindow))
           WINDISABLEUPDATE(pWindow)--;
       }
    else
       WINDISABLEUPDATE(pWindow)++;
} /*EnableWindowUpdate*/


void EnableWindowBorder( HWINDOW pWindow, 
                         WORD    enable )
{
  if (IsWindow(pWindow))
     {
      if (enable)
         WINFLAGS(pWindow) &= ~WFL_NOBORDER;
      else
         WINFLAGS(pWindow) |=  WFL_NOBORDER;

      CalcClientArea(pWindow);
     }
} /*EnableWindowBorder*/



# define MINIMIZED_WIDTH  (16)
# define MINIMIZED_HEIGHT ( 2)


void SaveRestoreArea(HWINDOW hWindow)
{
 WINRESTOREXPOS(hWindow)   = WINXPOS(hWindow);
 WINRESTOREYPOS(hWindow)   = WINYPOS(hWindow);
 WINRESTOREWIDTH(hWindow)  = WINWIDTH(hWindow);
 WINRESTOREHEIGHT(hWindow) = WINHEIGHT(hWindow);
}


void ZoomWindow( HWINDOW hWindow, 
                 HitType ht )
{
 if (IsWindow(hWindow))
    {
     switch (ht)
            {
             case HT_RESTORE:
               if (WINFLAGS(hWindow) & (WFL_MAXIMIZED | WFL_MINIMIZED))
                  {
                   ShowWindowNoUpdate(hWindow, FALSE);

                   WindowSetPosNoUpdate( hWindow,
                                         WINRESTOREXPOS(hWindow),
                                         WINRESTOREYPOS(hWindow) );

                   WINFLAGS(hWindow) &= ~(WFL_MAXIMIZED | WFL_MINIMIZED);

                   ResizeWindow( hWindow, 
                                 WINRESTOREWIDTH(hWindow),
                                 WINRESTOREHEIGHT(hWindow) );

                   ShowWindow(hWindow, TRUE);
                  }
               break;

             case HT_MAXIMIZE:
               if (! (WINFLAGS(hWindow) & WFL_MAXIMIZED))
                  {
                   if (! (WINFLAGS(hWindow) & WFL_MINIMIZED))
                      SaveRestoreArea(hWindow);

                   ShowWindowNoUpdate(hWindow, FALSE);

                   WindowSetPosNoUpdate( hWindow,
                                         WINRESTORERIGHT(hWindow) 
                                           - WINLOGICALWIDTH(hWindow) - 1,
                                         WINRESTOREYPOS(hWindow) );

                   WINFLAGS(hWindow) &= ~(WFL_MAXIMIZED | WFL_MINIMIZED);
                   WINFLAGS(hWindow) |= WFL_MAXIMIZED;

                   ResizeWindow( hWindow, 
                                 screenWidth  + 2,
                                 screenHeight + 1 );

                   ShowWindow(hWindow, TRUE);
                  }
               break;

             case HT_MINIMIZE:
               if (! (WINFLAGS(hWindow) & WFL_MINIMIZED))
                  {
                   if (! (WINFLAGS(hWindow) & WFL_MAXIMIZED))
                      SaveRestoreArea(hWindow);

                   ShowWindowNoUpdate(hWindow, FALSE);

                   WindowSetPosNoUpdate( hWindow,
                                         WINRESTORERIGHT(hWindow) 
                                           - MINIMIZED_WIDTH + 1,
                                         WINRESTOREYPOS(hWindow) );

                   WINFLAGS(hWindow) &= ~(WFL_MAXIMIZED | WFL_MINIMIZED);
                   WINFLAGS(hWindow) |= WFL_MINIMIZED;

                   /*
                      The size that consistutes 'minimized' is fairly 
                      arbitrary. We choose a size that shows no client
                      area (top and bottom borders are adjacent) but
                      which does allow the "Session X" part of the
                      caption to be displayed. I think this is reaonable
                      if not perfect.  The funny thing is that the user
                      *CAN* reduce the size of a so-called 'minimized'
                      window by manual resizing!
                   */

                   ResizeWindow( hWindow, 
                                 MINIMIZED_WIDTH,
                                 MINIMIZED_HEIGHT );

                   ShowWindow(hWindow, TRUE);
                  }
               break;
            }

    }
}



BOOL IsWindowBorderEnabled(HWINDOW hWindow)
{
 WORD isEnabled = FALSE;

 if (IsWindow(hWindow))
    isEnabled = ! (WINFLAGS(hWindow) & WFL_NOBORDER);

 return isEnabled;
}


void ClearWindowRect(HWINDOW pWindow, SHORT x, SHORT y, SHORT w, SHORT h)
{
 if (IsWindow(pWindow))
    if (     x < WINLOGICALWIDTH(pWindow)  
         &&  y < WINLOGICALHEIGHT(pWindow))
      {
        if (x + w > WINLOGICALWIDTH(pWindow))
          w = WINLOGICALWIDTH(pWindow)  - x;

        if (y + h > WINLOGICALHEIGHT(pWindow))
          h = WINLOGICALHEIGHT(pWindow) - y;

        if (w && h)
          {
            USHORT cell = (USHORT) WINBACKGROUNDCHAR (pWindow)
                           | ((USHORT) WINDEFATTR (pWindow) << 8);
            SHORT  yCtr;

            for (yCtr=y+h-1; yCtr>=y; yCtr--)
              {
                PUSHORT p = (PUSHORT)&WINCHARXY(pWindow, x, yCtr);
                SHORT  xCtr;

                for (xCtr=x+w-1; xCtr>=x; xCtr--)
                    *p++ = cell;
              }

            UpdateWindow(pWindow);
          }
      }
} /*ClearWindowRect*/


void ClearWindow(HWINDOW pWindow)
{
 if (IsWindow(pWindow))
    {
     ClearWindowRect( pWindow,
                      0,
                      0,
                      WINLOGICALWIDTH(pWindow),
                      WINLOGICALHEIGHT(pWindow) );

     WINLOGICALX(pWindow) = 0;
     WINLOGICALY(pWindow) = 0;

     WINCURX(pWindow) = 0;
     WINCURY(pWindow) = 0;

     UpdateWindow(pWindow);
    }
} /*ClearWindow*/


void SetEraseCharacter(HWINDOW w, char c)
{
 if (IsWindow(w))
    WINBACKGROUNDCHAR(w) = c;
}


void UpdateCursorDisplay(HWINDOW pWindow)
{
 int physX,
     physY,
     curX,
     curY,
     logX,
     logY;

 if (IsWindow(pWindow))
    if (pWindow == focuswindow)
       {
        curX = WINCURX(pWindow);
        curY = WINCURY(pWindow);
       
        logX = WINLOGICALX(pWindow);
        logY = WINLOGICALY(pWindow);
       
        physX = curX - logX + WINCLIENTXPOS(pWindow);
        physY = curY - logY + WINCLIENTYPOS(pWindow);
       
        if (    physY < 0  
             || physY > screenHeight
             || physY < WINCLIENTTOP(pWindow)
             || physY > WINCLIENTBOTTOM(pWindow) 
             || physX < 0  
             || physX > screenWidth
             || physX < WINCLIENTLEFT(pWindow)  
             || physX > WINCLIENTRIGHT(pWindow) )
           {
            physX = screenWidth;
            physY = screenHeight;
           }
       
        SetCursorPos(physY, physX);
       }
} /*UpdateCursorDisplay*/


void GetScreenMetrics(void)
{
 BYTE FAR * p;
 BYTE mode;

 BEGIN_ASM

   // Get current video mode

   mov ah,0x0f
   int 0x10
   mov byte ptr mode, al

 END_ASM

 if (mode >= 7  &&  mode <= 10)
    {
     forcedToMono = FALSE; // Already in mono...
     isMonoScreen = TRUE;
    }
 else
    {
     forcedToMono = forceMono;
     isMonoScreen = forceMono;
    }

 if (forcedToMono)
    {
     screenHeight     = 25;
     screenWidth      = 80;
     screenCharHeight = 8;

     mouseKind    = 0; /* Soft */
     isMonoScreen = TRUE;
    }
 else
    {
     p = (BYTE FAR *)0x0040004aL;
     screenWidth = *p;

     p = (BYTE FAR *)0x00400084L;
     screenHeight = (*p + 1);

     p = (BYTE FAR *)0x00400085L;
     screenCharHeight = *p;
    }

 if (screenWidth == 0)
    screenWidth = 80;

 if (screenHeight == 0)
    screenHeight = 25;

 if (screenCharHeight == 0)
    screenCharHeight = 8;

 if (isMonoScreen)
    {
     screenBuffer           = (PVOID)0xb0000000L;
//     wibbleFactor           = MONO_WIBBLE;
    }
 else
    {
     screenBuffer           = (PVOID)0xb8000000L;
//     wibbleFactor           = COLOUR_WIBBLE;
    }


 // AMcC : 12Dec92 : be a bit careful about when we set the default
 //                  colour attributes. This routine may be called
 //                  a number of times as textwin runs and we don't
 //                  really want to undo the user's colour settings.

 if (defaultWindowAttribute == 0)  // ie, if it is uninitialised
    if (isMonoScreen)
       defaultWindowAttribute = MONO_ATTRIBUTE;
    else
       defaultWindowAttribute = COLOUR_ATTRIBUTE;
}


# endif /* defined(TEXTWIN) */


/* eof: textwin.c */

