/* textwinc.c */

/*
   Text win command line stuff...
*/

/*
 *	$Id: textwinc.c 1.2 1997/06/20 11:29:04 Graham Exp $
 *	Jun 97	1.2	GeK	Added ip.h to fix warning
 */

#include <stdlib.h>
#include <stdio.h>
#include <conio.h>
#include <dir.h>
#include <dos.h>
#include <io.h>
#include <sys/stat.h>
#include <string.h>
#include <process.h>
#include <fcntl.h>
#include <alloc.h>
#include <stdarg.h>
#include <time.h>
#include <bios.h>
#include "config.h"
#include "global.h"
#include "files.h"
#include "mbuf.h"
#include "proc.h"
#include "ip.h"			/*  GeK - fix warning  */
#include "iface.h"
#include "internet.h"
#include "session.h"
#include "socket.h"
#include "cmdparse.h"
#include "textwin.h"
#include "textwinp.h"


int  doemulation __ARGS((int argc, char *argv[], void *p));


static int getwinid __ARGS((int argc, char* argv[], void * p));


static int
dotwnoop(argc,argv,p)
int argc;
char *argv[];
void *p;
{
 UNUSED_ARG(argc);
 UNUSED_ARG(argv);
 UNUSED_ARG(p);

 return 0;
}


static int
dotwdefcascade(argc,argv,p)
int argc;
char *argv[];
void *p;
{
 UNUSED_ARG(p);

 if (argc != 3)
    {
     tprintf( "cascade: %d %d\n", 
              cascadeRangeX,
              cascadeRangeY );
    }
 else
    {
     cascadeRangeX = max(0, atoi(argv[1]));
     cascadeRangeY = max(0, atoi(argv[2]));
    }

 return 0;
}

static int
dotwdefcharacter(argc,argv,p)
int argc;
char *argv[];
void *p;
{
 int which;
 char * chars;

 UNUSED_ARG(p);

 which = atoi(argv[1]);

 if (which & 1)
    chars = doubleFrameChars;
 else
    chars = singleFrameChars;

 if (which < 0  ||  which >= 22*2)
    tprintf("must be 0 - 43\n");
 else
    if (argc == 2)
       tprintf("char %d: %d\n", which, (BYTE)chars[which / 2]);
    else
       chars[which / 2] = (BYTE)atoi(argv[2]);

 return 0;
}


static int
dotwdefposition(argc,argv,p)
int argc;
char *argv[];
void *p;
{
 UNUSED_ARG(p);

 if (argc != 3)
    {
     tprintf( "def pos: %d %d\n", 
              defaultWindowX,
              defaultWindowY );
    }
 else
    {
     defaultWindowX = atoi(argv[1]);
     defaultWindowY = atoi(argv[2]);
    }

 return 0;
}

static int
dotwdefsize(argc,argv,p)
int argc;
char *argv[];
void *p;
{
 UNUSED_ARG(p);

 if (argc != 3)
    {
     tprintf( "def size: %d %d\n", 
              defaultWindowDX,
              defaultWindowDY );
    }
 else
    {
     defaultWindowDX = atoi(argv[1]);
     defaultWindowDY = atoi(argv[2]);
    }

 return 0;
}


struct cmds far TextwinDefaultCmds[] = {
    { "",          dotwnoop,         0, 0, NULLCHAR },
    { "cascade",   dotwdefcascade,   0, 0, NULLCHAR },
    { "character", dotwdefcharacter, 0, 2, "character <which> [<what>]" },
    { "position",  dotwdefposition,  0, 0, NULLCHAR },
    { "size",      dotwdefsize,      0, 0, NULLCHAR },
    { NULLCHAR,	   NULLFP,	     0, 0, NULLCHAR},
};


static int
dotwdefault(argc,argv,p)
int argc;
char *argv[];
void *p;
{
 return subcmd(TextwinDefaultCmds,argc,argv,p);
}



static int
dotwmouhard(argc,argv,p)
int argc;
char *argv[];
void *p;
{
 UNUSED_ARG(argc);
 UNUSED_ARG(argv);
 UNUSED_ARG(p);

 HideMouse();

 mouseKind = 0;

 ShowMouse();

 return 0;
}

static int
dotwmousoft(argc,argv,p)
int argc;
char *argv[];
void *p;
{
 UNUSED_ARG(argc);
 UNUSED_ARG(argv);
 UNUSED_ARG(p);

 HideMouse();

 mouseKind = 1;

 ShowMouse();

 return 0;
}

struct cmds far TextwinMouseCmds[] = {
    { "",          dotwnoop,      0, 0, NULLCHAR },
    { "hard",      dotwmouhard,   0, 0, NULLCHAR },
    { "soft",      dotwmousoft,   0, 0, NULLCHAR },
    { NULLCHAR,    NULLFP,        0, 0, NULLCHAR },
};


static int
dotwmouse(argc,argv,p)
int argc;
char *argv[];
void *p;
{
 char * kind;

 if (argc <= 1)
    {
     switch (mouseKind)
            {
             default:
             case 0: kind = "soft"; break;
             case 1: kind = "hard"; break;
            }

     tprintf("mouse: %s\n", kind);

     return 0;
    }
 else
    return subcmd(TextwinMouseCmds,argc,argv,p);
}


static int
getwinid(argc, argv, p)
int argc;
char *argv[];
void *p;
{
 int winid = 0;

 UNUSED_ARG(p);

 if (argc < 2)
    tprintf("Usage: %s <winid> ...\n", argv[0]);
 else
    {
     winid = atoi(argv[1]);

     if (winid <= 1) /* Don't allow access to the desktop window */
        winid = 0;
    }

 return winid;
}


static int
dotwborders(argc,argv,p)
int argc;
char *argv[];
void *p;
{
 UNUSED_ARG(p);

 setbool(&wantBorders, "borders", argc, argv);

 return 0;
}


static int
dotwcls(argc,argv,p)
int argc;
char *argv[];
void *p;
{
 int winid;
 HWINDOW h;

 winid = getwinid(argc, argv, p);

 if (winid)
    {
     h = FindWindowFromId(winid);

     ClearWindow(h);
    }

 return 0;
}

static int
dotwcolour(argc,argv,p)
int argc;
char *argv[];
void *p;
{
 int fg = (defaultWindowAttribute & 0x0F);
 int bg = (defaultWindowAttribute & 0xF0) >> 4;

 UNUSED_ARG(p);

 if (argc <= 1)
    tprintf("colour: fg %d bg %d\n", fg, bg);
 else
    {
     fg = atoi(argv[1]);
     bg = atoi(argv[2]);

     defaultWindowAttribute = ((bg & 0x0F) << 4) | (fg & 0x0F);
    }

 return 0;
}

static int
dotwfocus(argc,argv,p)
int argc;
char *argv[];
void *p;
{
 int winid;
 HWINDOW h;

 if (argc > 1)
    {
     winid = getwinid(argc, argv, p);
     
     if (winid)
        {
         h = FindWindowFromId(winid);
     
         SetWindowFocus(h);
        }
    }
 else 
    {
     h = GetWindowFocus();

     if (h == NULL)
        tprintf("No focus\n");
     else
        tprintf("Win %d has focus\n", WINID(h));
    }

 return 0;
}

static int
dotwhide(argc,argv,p)
int argc;
char *argv[];
void *p;
{
 int winid;
 HWINDOW h;

 winid = getwinid(argc, argv, p);

 if (winid)
    {
     h = FindWindowFromId(winid);

     ShowWindow(h, FALSE);
    }

 return 0;
}

static int
dotwhmouse(argc,argv,p)
int argc;
char *argv[];
void *p;
{
 int hardMouse = mouseKind == 0;

 UNUSED_ARG(p);

 setbool(&hardMouse, "hardmouse", argc, argv);

 if (mouseKind)
    {
     if (hardMouse)
        {
         HideMouse();

         mouseKind = hardMouse ? 0 : 1;

         ShowMouse();
        }
    }
 else /* !softMouse */
    {
     if (! hardMouse)
        {
         HideMouse();

         mouseKind = hardMouse ? 0 : 1;

         ShowMouse();
        }
    }

 return 0;
}

static int
dotwlist(argc,argv,p)
int argc;
char *argv[];
void *p;
{
 HWINDOW h;
 int timeout = 1000;
 PSZ caption;

 UNUSED_ARG(argc);
 UNUSED_ARG(argv);
 UNUSED_ARG(p);

 tprintf("Id Address  X   Y   wW  wH  tW  tH  Vis Fg Bg  Caption\n");

 h = GetWindowTop();

 while (h != NULL  &&  timeout--)
       {
        if (WINID(h) == 1)
           caption = "(Desktop)";
        else if (WINCAPTION(h) != NULL)
           caption = WINCAPTION(h);
        else
           caption = "(none)";

        tprintf( "%-2d %08lx %-3d %-3d %-3d %-3d %-3d %-3d  %-2d %-2d %-2d  %.30s\n",
                 WINID(h),
                 (long)h,
                 WINLEFT(h),
                 WINTOP(h),
                 WINWIDTH(h),
                 WINHEIGHT(h),
                 WINLOGICALWIDTH(h),
                 WINLOGICALHEIGHT(h),
                 (WINFLAGS(h) & WFL_HIDDEN) ? 0 : 1,
                 (WINDEFATTR(h) & 0x0F),
                 (WINDEFATTR(h) & 0xF0) >> 4,
                 caption );

        h = WINNEXTWINDOW(h);
       }

 return 0;
}

static int
dotwmove(argc,argv,p)
int argc;
char *argv[];
void *p;
{
 int winid;
 HWINDOW h;

 winid = getwinid(argc, argv, p);

 if (winid)
    {
     h = FindWindowFromId(winid);

     WindowSetPos(h, atoi(argv[2]), atoi(argv[3]));
    }

 return 0;
}

static int
dotwredraw(argc,argv,p)
int argc;
char *argv[];
void *p;
{
 int winid;
 HWINDOW h;

 if (argc > 1)
    {
     winid = getwinid(argc, argv, p);
     
     if (winid)
        {
         h = FindWindowFromId(winid);
     
         UpdateWindow(h);
        }
    }
 else
    UpdateWindowsBelowAndIncluding(GetWindowTop());

 return 0;
}

static int
dotwshow(argc,argv,p)
int argc;
char *argv[];
void *p;
{
 int winid;
 HWINDOW h;

 winid = getwinid(argc, argv, p);

 if (winid)
    {
     h = FindWindowFromId(winid);

     ShowWindow(h, TRUE);
    }

 return 0;
}

static int
dotwsize(argc,argv,p)
int argc;
char *argv[];
void *p;
{
 int winid;
 HWINDOW h;

 winid = getwinid(argc, argv, p);

 if (winid)
    {
     h = FindWindowFromId(winid);

     ResizeWindow(h, atoi(argv[2]), atoi(argv[3]));
    }

 return 0;
}

static int
dotwterm(argc,argv,p)
int argc;
char *argv[];
void *p;
{
 UNUSED_ARG(p);

 if (argc != 1  &&  argc != 3)
    tprintf("Usage: term <term_width> <term_height>\n");
 else if (argc == 1)
    tprintf( "Next win term size: %d %d\n",
             defaultLogicalWidth,
             defaultLogicalHeight );
 else if (argc == 3)
    {
     defaultLogicalWidth  = max(1, atoi(argv[1])); /* AMcC : 29Jul92    */
     defaultLogicalHeight = max(1, atoi(argv[2])); /* Removed max limit */
    }

 return 0;
}


static int
dotwscrmode(argc,argv,p)
int argc;
char *argv[];
void *p;
{
 UNUSED_ARG(p);

 if (argc != 2)
    {
     if (scrMode < 0)
        tprintf( "screen mode: %d (unchanged)\n",
                 -scrMode);
     else
        tprintf( "screen mode: %d\n", 
                 scrMode);
    }
 else
 if (bTextwinInitialized)
    tprintf("Only inside textwin.cfg\n");
 else
    {
     //
     // NOTE: This _ought_ to call setint(), which _ought_
     //       to be able to recognise and correctly respond
     //       to non-numeric input, especially '?'
     //

     scrMode = atoi(argv[1]);
    }

 return 0;
}


static int
dotwscrmono(argc,argv,p)
int argc;
char *argv[];
void *p;
{
 UNUSED_ARG(p);

 if ( (argc > 1)          &&
      (argv[1][0] != '?') &&
      bTextwinInitialized   )
    tprintf("Only inside textwin.cfg\n");
 else
    setbool(&forceMono, "mono", argc, argv);

 return 0;
}


static int
dotwscrrestore(argc,argv,p)
int argc;
char *argv[];
void *p;
{
 UNUSED_ARG(p);

 if ( (argc > 1)          &&
      (argv[1][0] != '?') &&
      bTextwinInitialized   )
    tprintf("Only inside textwin.cfg\n");
 else
    setbool(&scrRestore, "restore", argc, argv);

 return 0;
}


struct cmds far TextwinScreenCmds[] = {
    { "",        dotwnoop,       0, 0, NULLCHAR },
    { "mode",    dotwscrmode,    0, 0, NULLCHAR },
    { "mono",    dotwscrmono,    0, 0, NULLCHAR },
    { "restore", dotwscrrestore, 0, 0, NULLCHAR },
    { NULLCHAR,  NULLFP,         0, 0, NULLCHAR },
};


static int
dotwscreen(argc,argv,p)
int argc;
char *argv[];
void *p;
{
 return subcmd(TextwinScreenCmds,argc,argv,p);
}





/* Windowed Clock stuff */

static HWINDOW       far ClockWindow = NULL;
static struct proc * far ClockProcess = NULL;
static time_t        far startTime;

void twclockprocess __ARGS((int i,void *v1,void *v2));
int dotwclock __ARGS((int argc,char *argv[],void *p));


void
twclockprocess(i,v1,v2)
int i;
void *v1;
void *v2;
{
 time_t nowTime;
 struct tm * tblock;
 char str[40];
 time_t timeDiff;

 UNUSED_ARG(i);
 UNUSED_ARG(v1);
 UNUSED_ARG(v2);

 for (;;)
     {

      nowTime = time(NULL);
      tblock = localtime(&nowTime);

      strncpy(str, asctime(tblock), sizeof(str)-1);
      str[sizeof(str)-1] = '\0';
      if (str[strlen(str)-1] == '\n')
         str[strlen(str)-1] = '\0';
         
      PutDirtyStrAt(ClockWindow, 0, 0, str);

      timeDiff = nowTime - startTime;

      sprintf( str,
               "%dh %02dm %02ds",
               (int)(timeDiff / 3600L),
               (int)(timeDiff % 3600L) / 60,
               (int)(timeDiff % 60L) );

      PutDirtyStrAt(ClockWindow, 0, 1, str);

      UpdateDirtyArea(ClockWindow);

      alarm(5000);
      pwait(&ClockProcess);
     }
}


static int
dotwclockstart(argc,argv,p)
int argc;
char *argv[];
void *p;
{
 int x = 0,
     y = 0;

 UNUSED_ARG(p);

 if (! bTextwinInitialized)
    tprintf("'clock start' invalid in textwin.cfg\n");
 else
    if (ClockWindow == NULL)
       {
        if (argc > 1)
           x = atoi(argv[1]);
    
        if (argc > 1)
           y = atoi(argv[2]);
    
        ClockWindow = AllocTerminalWindow(x,y,24,2);
    
        SetWindowCaption(ClockWindow, "Clock");
    
        ClockProcess = newproc("twclock",1024,twclockprocess,0,NULL,NULL,0);
       }
    else
       tprintf("clock already running\n");

 return 0;
}

static int
dotwclockstop(argc,argv,p)
int argc;
char *argv[];
void *p;
{
 UNUSED_ARG(argc);
 UNUSED_ARG(argv);
 UNUSED_ARG(p);

 if (ClockWindow != NULL)
    {
     killproc(ClockProcess);
     ClockProcess = NULLPROC;

     FreeWindow(ClockWindow);
     ClockWindow = NULL;
    }
 else
    tprintf("clock not running\n");

 return 0;
}

static int
dotwclockreset(argc,argv,p)
int argc;
char *argv[];
void *p;
{
 UNUSED_ARG(p);

 startTime = time(NULL);

 if (argc > 1)
    startTime -= (time_t)atol(argv[1]);

 if (ClockProcess != NULLPROC)
    {
     ClearWindow(ClockWindow);
     psignal(&ClockProcess, 0);
    }

 return 0;
}


struct cmds far Clockcmds[] = {
    { "",         dotwnoop,       0, 0, NULLCHAR },
    { "reset",    dotwclockreset, 0, 0, NULLCHAR },
    { "start",    dotwclockstart, 0, 0, NULLCHAR },
    { "stop",     dotwclockstop,  0, 0, NULLCHAR },
    { NULLCHAR,   NULLFP,         0, 0, NULLCHAR },
};

static int
dotwclock(argc,argv,p)
int argc;
char *argv[];
void *p;
{
 return subcmd(Clockcmds,argc,argv,p);
}

static int
dotwparam(argc,argv,p)
int argc;
char *argv[];
void *p;
{
 int which = atoi(argv[1]);

 if (which <= 0  ||  which >= TW_PARAM_LAST)
    tprintf("Bad param\n");
 else
    if (argc <= 2)
       tprintf("param %d : %d\n", which, param[which]);
    else
       param[which] = atoi(argv[2]);

 return 0;
}



struct cmds far Textwincmds[] = {
    { "",         dotwnoop,    0, 0, NULLCHAR },
    { "borders",  dotwborders, 0, 0, NULLCHAR },
    { "clock",    dotwclock,   0, 0, NULLCHAR },
    { "cls",      dotwcls,     0, 2, "cls <winid>" },
    { "colour",   dotwcolour,  0, 0, NULLCHAR },
    { "default",  dotwdefault, 0, 1, "default [subcommand]" },
    { "emulation",doemulation, 0, 0, NULLCHAR },
    { "focus",    dotwfocus,   0, 2, "focus <winid>" },
    { "hardmouse",dotwhmouse,  0, 0, NULLCHAR },
    { "hide",     dotwhide,    0, 2, "hide <winid>" },
    { "key",      dotwkey,     0, 2, "key <which> [<ch1> [<ch2>]]" },
    { "list",     dotwlist,    0, 0, NULLCHAR },
    { "mouse",    dotwmouse,   0, 0, NULLCHAR },
    { "move",     dotwmove,    0, 4, "move <winid> <x> <y>" },
    { "param",    dotwparam,   0, 2, "param <which> [<what>]" },
    { "redraw",   dotwredraw,  0, 0, NULLCHAR },
    { "screen",   dotwscreen,  0, 2, "screen [subcommand]" },
    { "show",     dotwshow,    0, 2, "show <winid>" },
    { "size",     dotwsize,    0, 4, "size <winid> <dx> <dy>" },
    { "terminal", dotwterm,    0, 0, NULLCHAR },
    { NULLCHAR,   NULLFP,      0, 0, NULLCHAR },
};


int
dotextwin(argc,argv,p)
int argc;
char *argv[];
void *p;
{
 return subcmd(Textwincmds,argc,argv,p);
}



void ReadTextwinConfigFile(void)
{
 /*
    Read in the textwin init file.
 */

 FILE * fp;
 char * inbuf;
 char * intmp;
 char * ConfigFileName = rootdircat("textwin.cfg");


 startTime = time(NULL);


 fp = fopen(ConfigFileName, READ_TEXT);

 if (fp != NULLFILE)
    {
	 inbuf = mallocw(BUFSIZ);
	 intmp = mallocw(BUFSIZ);

	 while (fgets(inbuf,BUFSIZ,fp) != NULLCHAR)
           {
			strcpy(intmp,inbuf);

			cmdparse(Textwincmds,inbuf,NULL);
		   }

	 fclose(fp);
	 free(inbuf);
	 free(intmp);
	}
}


/* eof: textwinc.c */
