/* textwind.c */

/*
 *	$Id: textwind.c 1.2 1997/06/20 11:30:20 Graham Exp $
 *	Mar 97	1.2	GeK	Swapped while(1) for for(;;) to fix BC4 warning
 */

#include <stdlib.h>
#include <stdio.h>

#include "config.h"
#include "global.h"

#include "textwin.h"
#include "textwinp.h"


# ifdef TEXTWIN

/*
   This "LineHistory" stuff is a very crude debugging mechanism which
   shows us recent execution history in terms of which lines have been
   executed. Its a bit tacky but can be useful.
*/



# define MAX_LINE_HISTORY (16)

typedef struct tagLineHistory
{
 int next;

 WORD   line[MAX_LINE_HISTORY];
 char * file[MAX_LINE_HISTORY];
} LineHistory;

LineHistory far lineHistory = {0};

void AddLineHistory(WORD line, char * fname)
{
 lineHistory.line[lineHistory.next] = line;
 lineHistory.file[lineHistory.next] = fname;

 lineHistory.next++;

 lineHistory.next %= MAX_LINE_HISTORY;
}

void PrintLineHistory(void)
{
 int inx = lineHistory.next - 1;
 char * fname;

 if (inx < 0)
    inx = MAX_LINE_HISTORY - 1;

 printf("DANGER! REBOOT! MEMORY TRAMPLING ERROR DETECTED!\n");

 printf("Exec history:\n");

 while (    lineHistory.line[inx] != 0
         && inx != lineHistory.next )
       {
        fname = lineHistory.file[inx];

        if (fname == NULL)
           fname = "?File?";

        printf("%s(%d) ", fname, lineHistory.line[inx]);

        inx--;

        if (inx < 0)
           inx = MAX_LINE_HISTORY - 1;
       }

 printf("\n");

 /*  GeK - BC4 understands for() better; compiles to the same asm  */
 for (;;) /* loop until the user reboots! */
       ;
}



/*
   The amc_malloc() and amc_free() routines, together with the
   data structures below, attempt to watch for memory trampling
   either side of an allocated data block.

   We allocate a "few" (not just one) longs which contain known
   values when the block is active and when it has been freed.
   In this way, we can watch for code that tramples in either
   direction beyond the end/start of the memory block and we can
   also see if we're accidentally reusing a freed block.

   Also, having a large guard area like this improves the chance
   of detecting really wild writes... sort of like looking for
   neutrinos with bigger tubs of water...
*/


# define AMC_MAGIC_HEAD 0xFEEDF00DL
# define AMC_MAGIC_TAIL 0xFACEDEAFL

# define AMC_FREED_MAGIC_HEAD 0xFADECEDEL
# define AMC_FREED_MAGIC_TAIL 0xDEADD0D0L


# ifdef AMC_MEMORY_PARANOIA
#  define QTY_MAGIC (5) 
# else
#  define QTY_MAGIC (1) 
# endif


typedef struct tagAmcMallocHead FAR * LPAmcMallocHead;
typedef struct tagAmcMallocHead
{
 long magicHead[QTY_MAGIC];
 unsigned int size;
} AmcMallocHead;


typedef struct tagAmcMallocTail FAR * LPAmcMallocTail;
typedef struct tagAmcMallocTail
{
 long magicTail[QTY_MAGIC];
} AmcMallocTail;


# define AMC_EXTRA (sizeof(AmcMallocHead) + sizeof(AmcMallocTail))


PVOID amc_malloc(unsigned int size)
{
 PBYTE ptr = NULL;
 LPAmcMallocHead h;
 LPAmcMallocTail t;
 int ctr;

 ptr = mallocw(size + AMC_EXTRA);

 if (ptr == NULL)
    {
     printf("mallocw returned NULL\n");
     exit(1);
    }
 
 h = (LPAmcMallocHead)(ptr);
 t = (LPAmcMallocTail)(ptr + sizeof(AmcMallocHead) + size);

 h->size = size;
 for ( ctr = 0;
       ctr < QTY_MAGIC;
       ctr++ )
     {
      h->magicHead[ctr] = AMC_MAGIC_HEAD;
      t->magicTail[ctr] = AMC_MAGIC_TAIL;
     }

 return (ptr + sizeof(AmcMallocHead));
}

PBYTE amc_check(PVOID ptr, int atLine, char * file)
{
 PBYTE p = (PBYTE)ptr;
 int ctr;
 LPAmcMallocHead h;
 LPAmcMallocTail t;
 BOOL bFailure = FALSE;

 if (ptr == NULL)
    {
     printf("amc: NULL pointer\n");
     exit(1);
    }

 p -= sizeof(AmcMallocHead);

 h = (LPAmcMallocHead)(p);
 t = (LPAmcMallocTail)(p + h->size + sizeof(AmcMallocHead));


 for ( ctr = 0;
       ctr < QTY_MAGIC;
       ctr++ )
     {
      if (h->magicHead[ctr] == AMC_FREED_MAGIC_HEAD)
         {
          printf("amc: file %s line %d : freed magic head #%d\n", file, atLine, ctr);
          bFailure = TRUE;
         }
     }


 for ( ctr = 0;
       ctr < QTY_MAGIC;
       ctr++ )
     {
      if (t->magicTail[ctr] == AMC_FREED_MAGIC_TAIL)
         {
          printf("amc: file %s line %d : freed magic tail #%d\n", file, atLine, ctr);
          bFailure = TRUE;;
         }
     }

 for ( ctr = 0;
       ctr < QTY_MAGIC;
       ctr++ )
     {
      if (h->magicHead[ctr] != AMC_MAGIC_HEAD)
         {
          printf("amc: file %s line %d : crap magic head #%d\n", file, atLine, ctr);
          bFailure = TRUE;
         }
     }


 for ( ctr = 0;
       ctr < QTY_MAGIC;
       ctr++ )
     {
      if (t->magicTail[ctr] != AMC_MAGIC_TAIL)
         {
          printf("amc: file %s line %d : crap magic tail #%d\n", file, atLine, ctr);
          bFailure = TRUE;;
         }
     }

 if (bFailure)
    {
     PrintLineHistory();
     exit(1);
    }

 return p;
}

void amc_free(PVOID ptr)
{
 PBYTE p = amc_check(ptr, __LINE__, __FILE__);
 LPAmcMallocHead h;
 LPAmcMallocTail t;
 int ctr;

 h = (LPAmcMallocHead)(p);
 t = (LPAmcMallocTail)(p + h->size + sizeof(AmcMallocHead));

 for ( ctr = 0;
       ctr < QTY_MAGIC;
       ctr++ )
     {
      h->magicHead[ctr] = AMC_FREED_MAGIC_HEAD;
      t->magicTail[ctr] = AMC_FREED_MAGIC_TAIL;
     }

 h->size = 0;

 free(p);
}

# ifdef AMC_MEMORY_PARANOIA


static void amc_prim_checkwindow(PW p, int line, char * fname)
{
 amc_check((PVOID)p, line, fname);
 amc_check((PVOID)WINLOGICALBUFFER(p), line, fname);
 amc_check((PVOID)WINVISFLAGS(p), line, fname);
 amc_check((PVOID)WINVISBUFFER(p), line, fname);
}

BOOL amc_checkwindow(PW p, int line, char * fname)
{
 BOOL ok = TRUE;

 AddLineHistory(line, fname);

 amc_prim_checkwindow(p, line, fname);

 return ok;
}

BOOL amc_checkall(PW pw, char * file, int line)
{
 BOOL ok = TRUE;
 PW p = (PW)topwindow;
 int timeout = 100;

 AddLineHistory(line, file);

 amc_prim_checkwindow(pw, line, file);

 while (p != NULL  &&  timeout-- > 0)
       {
        amc_prim_checkwindow(p, line, file);

        p = WINNEXTWINDOW(p);
       }

 return ok;
}


# endif


# endif /* TEXTWIN */

/* eof: textwind.c */
